# ------------------------------------------------------------------------------
# Project: IFE_Surrogate
# Authors: Tobias Leitgeb, Julian Tischler
# CD Lab 2025
# ------------------------------------------------------------------------------
from ife_surrogate.gp.trainers import Trainer
from ife_surrogate.gp.models import GPModel

from jax import vmap, jit
import typing
import pyswarms as ps
from jax.flatten_util import ravel_pytree
import numpy as np
from jaxtyping import Array

import jax.numpy as jnp
from jax import jit, vmap
from jax.flatten_util import ravel_pytree
import numpy as np
import pyswarms as ps
from flax.core.frozen_dict import FrozenDict
import jax.tree_util as tree_util

TypeGPModel = typing.TypeVar("GPModel", bound=GPModel)


class SwarmTrainer(Trainer):
    r"""
    Trainer class that optimizes model parameters using Particle Swarm Optimization (PSO) via pyswarms[1]

    This class extends a generic `Trainer` and implements a global optimization
    method for training models, particularly useful when gradients are unavailable
    or the loss landscape is highly non-convex.

    Args:
        swarm_settings (dict, optional): Dictionary of PSO hyperparameters:
            - `c1` (float): Cognitive parameter (default: 0.5)
            - `c2` (float): Social parameter (default: 0.3)
            - `w` (float): Inertia weight (default: 0.9)
        number_iterations (int, optional): Number of iterations for the PSO algorithm per restart (default: 100).
        number_particles (int, optional): Number of particles in the swarm (default: 20).
        number_restarts (int, optional): Number of independent PSO runs to avoid local minima (default: 1).
        bounds (dict[str: (float, float)], optional): Lower and upper bounds for each parameter. 
            Not defined parameter bounds are automatically inferred from kernel priors.
        **kwargs: Additional keyword arguments passed to the base `Trainer` class.

    Attributes:
        swarm_settings (dict): Hyperparameters controlling particle behavior in PSO.
        number_iterations (int): Iterations per PSO run.
        number_particles (int): Number of particles in the swarm.
        number_restarts (int): Number of independent PSO runs.
        bounds (dict[str: (float, float)]): Parameter bounds.
        model: The model to be optimized.
        verbose (bool): Flag to control printing of optimization progress (inherited from `Trainer`).

    Methods:
        info(): Prints current PSO settings and bounds.
        train(model): Optimizes the provided model's kernel parameters using PSO and returns
                      the best parameters and optionally the optimization history.

    Example:
        >>> trainer = SwarmTrainer(number_particles=30, number_iterations=200)
        >>> best_run, history = trainer.train(model)
    
    [1] 'https://github.com/ljvmiranda921/pyswarms'
    """

    def __init__(self, 
                 swarm_settings={"c1":1.5, "c2":1.5, "w":0.5},
                 number_iterations: int = 100,
                 number_particles: int = 20,
                 number_restarts: int = 1,
                 bounds: dict = None,
                 **kwargs):
        super().__init__(**kwargs)
        self.swarm_settings = swarm_settings
        self.number_iterations = number_iterations
        self.number_particles = number_particles
        self.number_restarts = number_restarts
        self.bounds = bounds if bounds else {}

    def info(self):
        print("SwarmTrainer settings: ")
        print("\tswarm_settings: ", self.swarm_settings)
        print("\tnumber_iterations: ", self.number_iterations)
        print("\tnumber_particles: ", self.number_particles)
        print("\tnumber_restarts: ", self.number_restarts)
        print("\tbounds: ", self.bounds)

    def train(self, model):
        self.model = model
        kernel = model.kernel
        nlml = self.nlml

        params_template = kernel.get_params()
        flat_template, unravel_fn = ravel_pytree(params_template)
        dim = len(flat_template)

        ## Generate bounds automatically, if not defined
        priors = kernel.get_priors()
        keys = params_template.keys()
        for name in priors.keys():
            if name in keys and name not in self.bounds.keys():
                args = priors[name].get_args()
                low = args.get("low")
                high = args.get("high")
                self.bounds[name] = [max(0.0, low), max(0.0, high)]

        _bounds = ([], [])
        for name, __bounds in self.bounds.items():
            if name in params_template.keys():
                _lb = [__bounds[0]] * len(params_template[name])
                _ub = [__bounds[1]] * len(params_template[name])
                _bounds[0].extend(_lb)
                _bounds[1].extend(_ub)

        ## function for raveling and unraveling the hyperparameter dictionary of the kernel
        params_template = kernel.get_params()
        raveled_template, unravel_fn = ravel_pytree(params_template)        
        
        # def single_particle_loss_(flat_params: Array) -> float:
        #     #params = array_to_dict(params)
        #     params = unravel_fn(flat_params)
        #     kernel.update_params(params)
        #     return nlml(kernel)

        def single_particle_loss_(flat_params: Array) -> float:
        
            params = unravel_fn(flat_params)
            kernel.update_params(params)
            # CRITICAL FIX: Use .replace() instead of .update_params()
            # Flax dataclasses are immutable. .replace() creates a new instance safely.
            # This bypasses the in-place mutation bug in Kernel.update_params.
            updated_kernel = kernel.replace(**params)
            
            # Pass the NEW kernel instance to the loss function
            return nlml(updated_kernel)
        
        single_particle_loss = jit(vmap(single_particle_loss_))
        # single_particle_loss = single_particle_loss_
        
        history = {}
        for r in range(self.number_restarts):
            optimizer = ps.single.GlobalBestPSO(
                n_particles=self.number_particles,
                dimensions=dim,
                options=self.swarm_settings,
                bounds=_bounds
            )

            best_cost, best_pos = optimizer.optimize(
                single_particle_loss,
                iters=self.number_iterations,
                verbose=True
            )
            
            best_params = unravel_fn(best_pos)
            kernel.update_params(best_params)

            history[f"run_{r}"] = {
                "params": best_params,
                "loss": best_cost
            }
            if self.verbose:
                print(f"Run {r}: Best loss = {best_cost:.5f}")

        ## Remove nans, there should not be any though
        history = {k: v for k, v in history.items() if not np.isnan(v["loss"])}
        best_key = min(history, key=lambda k: history[k]["loss"])

        if self.verbose:
            print("Overall best run:", best_key, "loss:", history[best_key]["loss"])

        best_run = history[best_key]
        return (best_run, history) if self.save_history else (best_run, {})
    
    # def train(self, model):
    #     print("New train fn")
    #     self.model = model
    #     kernel = model.kernel
    #     nlml = self.nlml

    #     # 1. Get the template and flattening function
    #     params_template = kernel.get_params()
    #     flat_template, unravel_fn = ravel_pytree(params_template)
    #     dim = len(flat_template)

    #     # 2. Infer bounds if not present
    #     priors = kernel.get_priors()
    #     param_keys = params_template.keys() if hasattr(params_template, 'keys') else []
        
    #     for name in priors.keys():
    #         if name in param_keys and name not in self.bounds:
    #             args = priors[name].get_args()
    #             low = args.get("low")
    #             high = args.get("high")
    #             l_val = max(0.0, low) if low is not None else 0.0
    #             h_val = max(0.0, high) if high is not None else 1e5 
    #             self.bounds[name] = [l_val, h_val]

    #     # 3. Construct _bounds vector aligned with flat_template
    #     # FIX: We construct bounds by mapping over the parameter structure, not the bounds dict
        
    #     def get_bounds_for_leaf(path, leaf_val):
    #         # path is a tuple of keys/indices. path[0] is usually the param name.
    #         if not path:
    #             return (-jnp.inf, jnp.inf)
            
    #         # Extract key name safely
    #         key = str(path[0].key) if hasattr(path[0], 'key') else str(path[0])

    #         if key in self.bounds:
    #             b_low, b_high = self.bounds[key]
    #             # Ensure bounds match the shape of the parameter array
    #             return (jnp.full(leaf_val.shape, b_low), jnp.full(leaf_val.shape, b_high))
    #         else:
    #             return (jnp.full(leaf_val.shape, -np.inf), jnp.full(leaf_val.shape, np.inf))

    #     # Generate a PyTree of bounds matching the structure of params_template
    #     bounds_tree = tree_util.tree_map_with_path(get_bounds_for_leaf, params_template)
        
    #     # Flatten the bounds tree using the same unravel logic as the parameters
    #     flat_lb, _ = ravel_pytree(tree_util.tree_map(lambda x: x[0], bounds_tree))
    #     flat_ub, _ = ravel_pytree(tree_util.tree_map(lambda x: x[1], bounds_tree))
        
    #     _bounds = (np.array(flat_lb), np.array(flat_ub))

    #     # 4. Define Loss Function
    #     def single_particle_loss_(flat_params):
    #         params = unravel_fn(flat_params)
            
    #         # CRITICAL FIX: Use .replace() instead of .update_params()
    #         # Flax dataclasses are immutable. .replace() creates a new instance safely.
    #         # This bypasses the in-place mutation bug in Kernel.update_params.
    #         # We also ensure params is strictly a dict for unpacking.
    #         if not isinstance(params, dict):
    #              # This handles cases where unravel_fn might return a different PyTree structure
    #              # though ravel_pytree on a dict template should return a dict.
    #              pass

    #         updated_kernel = kernel.replace(**params)
            
    #         # Pass the NEW kernel instance to the loss function
    #         return nlml(updated_kernel)
            
    #     # JIT and VMAP for efficiency
    #     single_particle_loss = jit(vmap(single_particle_loss_))
        
    #     history = {}
    #     for r in range(self.number_restarts):
    #         optimizer = ps.single.GlobalBestPSO(
    #             n_particles=self.number_particles,
    #             dimensions=dim,
    #             options=self.swarm_settings,
    #             bounds=_bounds
    #         )

    #         # optimize() takes a function that accepts (N_particles, dimensions)
    #         best_cost, best_pos = optimizer.optimize(
    #             single_particle_loss,
    #             iters=self.number_iterations,
    #             verbose=self.verbose
    #         )

    #         best_params = unravel_fn(best_pos)
            
    #         history[f"run_{r}"] = {
    #             "params": best_params,
    #             "loss": best_cost
    #         }
    #         if self.verbose:
    #             print(f"Run {r}: Best loss = {best_cost:.5f}")

    #     # Filter out failed runs (NaNs)
    #     history = {k: v for k, v in history.items() if not np.isnan(v["loss"])}
        
    #     if not history:
    #          print("Optimization failed: All runs resulted in NaN loss.")
    #          return params_template, {}

    #     best_key = min(history, key=lambda k: history[k]["loss"])

    #     if self.verbose:
    #         print("Overall best run:", best_key, "loss:", history[best_key]["loss"])

    #     best_run = history[best_key]
        
    #     # We return the best parameters. The caller is responsible for applying them 
    #     # to the model if desired, as we cannot mutate the frozen Kernel here.
    #     return best_run, history if self.save_history else best_run


# ------------------------------------------------------------------------------
# Project: IFE_Surrogate
# Authors: Tobias Leitgeb, Julian Tischler
# CD Lab 2025
# ------------------------------------------------------------------------------
from jaxtyping import Array
from typing import Callable, Dict
import jax.numpy as jnp
from .kernel import Kernel
from flax import struct



@struct.dataclass
class Kriging(Kernel):
    r"""
    Kriging kernel (a generalized RBF kernel with a power term).

    Computes the covariance between two inputs x1 and x2 as:
    
    .. math::
        
        k_x(\mathbf x, \mathbf{x'}; \mathbf \theta) = \exp{\left(-\frac{1}{2}\sum_{i=1}^d \frac{1}{l_i^{2}} |x_{i} - x'_{i}|^{\kappa_i}\right)}
    
    where:  
    - `lengthscale` controls the smoothness along each input dimension.
    - `power` exponent of the absolute distance .

    Args:
        lengthscale (Array): Positive scaling factors for each input dimension.
        power (Array): Power applied to the distance; must be > 0.
        priors (Dict[str, Callable], optional): Optional priors for hyperparameters.
    """
    lengthscale: Array 
    power: Array
    

    def evaluate(self, x1: Array, x2: Array) -> Array:
        """
        Evaluate the Kriging kernel between x1 and x2.

        Args:
            x1 (Array): Inputs of shape (n_samples_1, n_features).
            x2 (Array): Inputs of shape (n_samples_2, n_features).

        Returns:
            Array: Kernel matrix of shape (n_samples_1, n_samples_2).
        """
        sq_dist = jnp.sum(jnp.abs(x1[:, None] - x2)**self.power * self.lengthscale**-2, axis=-1)
        return jnp.exp(-0.5 * sq_dist)

    









# ------------------------------------------------------------------------------
# Project: IFE_Surrogate
# Authors: Tobias Leitgeb, Julian Tischler
# CD Lab 2025
# ------------------------------------------------------------------------------
from .likelihood import nmll_wideband, nlml_scalar, nmll_kron
from .inference import predictive_mean_var_wideband, predictive_mean_var_scalar, predictive_mean_var_kron

__all__ = [ 
    "nmll_wideband", 
    "nlml_scalar",
    "nmll_kron",
    "predictive_mean_var_wideband",
    "predictive_mean_var_scalar",
    "predictive_mean_var_kron"
    ]


# ------------------------------------------------------------------------------
# Project: IFE_Surrogate
# Authors: Tobias Leitgeb, Julian Tischler
# CD Lab 2025
# ------------------------------------------------------------------------------
from abc import ABC, abstractmethod
import typing
from abc import ABC, abstractmethod
from typing import Callable, Dict, Tuple, Optional
from functools import partial
from ..models import GPModel


TypeGPModel = typing.TypeVar("GPModel", bound=GPModel)


class Trainer(ABC):
    def __init__(
        self,
        sample_parameters: bool = True,
        verbose: bool = False,
        save_history: bool = False
    ):
        self.sample_parameters = sample_parameters
        self.verbose = verbose
        self.save_history = save_history
        self._model = None

    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, new_model):
        self._model = new_model
        likelihood = self.model.likelihood
        if "sigma_sq" in self.model.get_attributes().keys():
            self.nlml = partial(likelihood, self.model.X, self.model.Y, self.model.sigma_sq, self.model.jitter)
        else:
            self.nlml = partial(likelihood, self.model.X, self.model.Y, self.model.jitter)


    def loss_fn(self, updated_params: Dict) -> float:
        return self.nlml(self.model.kernel)

    @abstractmethod
    def train(self, *args, **kwargs) -> Tuple[Dict, Optional[Dict]]:
        """Run optimization and return (best_run, history)."""
        pass


# ------------------------------------------------------------------------------
# Project: IFE_Surrogate
# Authors: Tobias Leitgeb, Julian Tischler
# CD Lab 2025
# ------------------------------------------------------------------------------
from ife_surrogate.gp.models.gp_model import GPModel
from ife_surrogate.gp.train import train, train_scipy
from ife_surrogate.gp.likelihood import nlml_scalar
from ife_surrogate.gp.inference import predictive_mean_var_scalar
from ife_surrogate.gp.kernels import Kernel
from ife_surrogate.utils.decorators import experimental

from jaxtyping import Key, Array, Int, Bool
from typing import Tuple, Dict
import typing
import jax.numpy as jnp
import numpyro
import optax
import numpyro.distributions as dist
from numpyro.infer import MCMC, NUTS


TypeKernel = typing.TypeVar("Kernel", bound=Kernel)


class ScalarGP(GPModel):
    r"""
    Scalar Gaussian Process (GP) model for scalar regression tasks.

    .. math::
        f \thicksim \mathcal{GP}(0, k(x, x'))

    Parameters
    ----------
    X : Array
        Training input data of shape (n_samples, n_features).
    Y : Array
        Training output data of shape (n_samples, 1)
    kernel : Kernel
        Instance of a kernel class defining the covariance structure.
    """

    def __init__(self, X: Array, Y: Array, kernel: TypeKernel):
        assert X.shape[0] == Y.shape[0] , "X and Y should have the same number of samples"
        assert Y.shape[1] == 1, "Y should be a matrix with shape (N, 1) N: #samples"
        super().__init__(kernel, X, Y)
        self.likelihood = nlml_scalar
        self.jitter = 1e-6
       
    def train(
        self, 
        key: Key,
        optim_dictionary: Dict={"opt": optax.adam, "settings": {"learning_rate": 1e-2}}, 
        n_steps: Int =1000, 
        n_restarts: int=1, 
        save_history: Bool=False, 
        verbose: Bool=False
    ) -> Tuple:
        """
        Optimize kernel parameters by minimizing the negative marginal log-likelihood using gradient-based optimizers.

        Parameters
        ----------
        key : Key
            Random key for parameter initialization and restarts.
        optim_dictionary : Dict, optional
            Dictionary containing the optimizer class (`"opt"`) and its settings (`"settings"`).
        num_steps : int, optional
            Number of optimization steps for each restart.
        number_restarts : int, optional
            Number of independent optimization restarts with different initializations.
        save_history : bool, optional
            If True, saves the full optimization trajectory (parameter history).
        verbose : bool, optional
            If True, prints optimization progress.

        Returns
        -------
        Tuple
            Optimized parameters and, if requested, parameter history.
        """
        
        self.optimized_parameters, self.parameter_history = train(
            self, key, optim_dictionary, n_steps, n_restarts, save_history, verbose
        )
        self.kernel.update_params(self.optimized_parameters["params"])

        pass

    def train_scipy(self, 
        key: Key, 
        number_restarts: int=1, 
        opt_algorithm: Dict={"method":"L-BFGS-B"},
        save_history: Bool=False, 
        verbose: Bool=True
    ) -> Tuple:
        """
        Optimize kernel parameters using a scipy-based optimizer (e.g., L-BFGS-B).

        Parameters
        ----------
        key : Key
            Random key for parameter initialization and restarts.
        number_restarts : int, optional
            Number of independent optimization restarts with different initializations.
        opt_algorithm : Dict, optional
            Dictionary specifying the optimization method and settings (e.g., `"method": "L-BFGS-B"`).
        save_history : bool, optional
            If True, saves the full optimization trajectory.
        verbose : bool, optional
            If True, prints optimization progress.

        Returns
        -------
        Tuple
            Optimized parameters and, if requested, parameter history.
        """
        self.optimized_parameters, self.parameter_history = train_scipy(
            self, key, number_restarts, opt_algorithm, save_history, verbose
        )
        self.kernel.update_params(self.optimized_parameters["params"])
    
    def predict(self, X_test: Array) -> Tuple[Array, Array]:
        """
        Predict the mean and variance for the given test inputs.

        Parameters
        ----------
        X_test : Array
            Test input data of shape (n_test_samples, n_features).

        Returns
        -------
        Tuple[Array, Array]
            Mean predictions and variances for each output dimension.
            Shapes: (n_test_samples, n_outputs)
        """ 
        X_test = jnp.atleast_2d(X_test)
        mean, var = (predictive_mean_var_scalar)(
            self.X, self.Y, self.kernel, self.jitter, X_test 
        )
        return mean, var
   
    def log_marginal_likelihood(self):
        """
        Compute the (negative) log marginal likelihood of the current model.

        Returns
        -------
        float
            The negative log marginal likelihood value.
        """
        return self.likelihood(self.X, self.Y, self.jitter, self.kernel)

    def sample_posterior(self, key: Key, X: Array, n_samples: Int) -> Array:
        """
        Sample from the posterior distribution at new input locations.

        Parameters
        ----------
        key : Key
            Random key for sampling.
        X : Array
            Input locations where to sample the posterior.
        n_samples : int
            Number of posterior samples to draw.

        Returns
        -------
        Array
            Posterior samples of shape (n_samples, n_test_samples, n_outputs).
        """
        return NotImplementedError
    
    def sample_prior(self, key: Key, X: Array, n_samples: Int) -> Array:
            """
            Sample from the prior distribution at new input locations.

            Parameters
            ----------
            key : Key
                Random key for sampling.
            X : Array
                Input locations where to sample the prior.
            n_samples : int
                Number of prior samples to draw.

            Returns
            -------
            Array
                Prior samples of shape (n_samples, n_test_samples, n_outputs).
            """
            return NotImplementedError

    def save(self, path: str):
        """
        Save the model parameters to a file.

        Parameters
        ----------
        path : str
            Path where the model parameters should be saved.
        """
        return NotImplementedError


@experimental
class ScalarGPBaysian(GPModel):
    r"""
    Fully baysian Scalar Gaussian Process (GP) model for scalar regression tasks.

    .. math::
        f \thicksim \mathcal{GP}(0, k(x, x'))

    Parameters
    ----------
    X : Array
        Training input data of shape (n_samples, n_features).
    Y : Array
        Training output data of shape (n_samples, 1)
    kernel : Kernel
        Instance of a kernel class defining the covariance structure.
    """

    def __init__(self, X: Array, Y: Array, kernel: TypeKernel):
        assert X.shape[0] == Y.shape[0] , "X and Y should have the same number of samples"
        assert Y.shape[1] == 1, "Y should be a matrix with shape (N, 1) N: #samples"
        super().__init__(kernel, X, Y)
        self.jitter = 1e-6
        self.likelihood = nlml_scalar
       
    def model_forward(self,):
        r"""
        Forward pass of the model required for Bayesian inference with NumPyro.

        This function defines the joint distribution over the observations and the kernel hyperparameters. 

        .. math:: 
        
             p(\mathbf{y} \mid \mathbf{X}, \boldsymbol{\theta}) p(\boldsymbol{\theta})

        By specifying priors over hyperparameters and a Gaussian Process likelihood, 
        we enable NumPyro to perform inference and approximate the posterior over kernel parameters:

        .. math::

            p(\boldsymbol{\theta} \mid \mathcal{D}) \propto 
            p(\mathbf{y} \mid \mathbf{X}, \boldsymbol{\theta}) \, p(\boldsymbol{\theta})

        where:

        - :math:`p(\boldsymbol{\theta})` are the user-defined priors over kernel parameters,
        - :math:`p(\mathbf{y} \mid \mathbf{X}, \boldsymbol{\theta})` is the GP marginal likelihood:

        .. math::

            \mathbf{y} \sim \mathcal{N}(\mathbf{0}, K_{\boldsymbol{\theta}} + \sigma^2 I)

        The kernel matrix :math:`K_{\boldsymbol{\theta}}` is computed using the sampled 
        hyperparameters. 
        """
        parameters = self.kernel.get_params() #dictionary of all hyperparamers
        hyperparameters = parameters.keys() #list of names of hyperparameters
        prior_functions = self.kernel.get_priors() #dictionary of all prior functions

        for hyperparam_name in hyperparameters:
            n_features = parameters[hyperparam_name].shape #extracts the dimension of the hperparam. e.g ard kernel
            sampled_hyperparam = numpyro.sample(
                hyperparam_name, prior_functions[hyperparam_name].expand(n_features)
            )
            #new sampled hyperparam is put into the dictionary
            parameters[hyperparam_name] = sampled_hyperparam.copy()
            self.kernel.update_params(parameters)#kernel is updated

        K = self.kernel(self.X, self.X) + self.jitter * jnp.eye(self.X.shape[0])

        numpyro.sample(
            "Y",
            dist.MultivariateNormal(loc=jnp.zeros(self.X.shape[0]), covariance_matrix=K),
            obs=self.Y,
        )

    def predict(self, samples, X_test: Array) -> Tuple[Array, Array]:
        """
        Predict the mean and variance for the given test inputs.

        Parameters
        ----------
        X_test : Array
            Test input data of shape (n_test_samples, n_features).

        Returns
        -------
        Tuple[Array, Array]
            Mean predictions and variances for each output dimension.
            Shapes: (n_test_samples, n_outputs)
        """ 
        self.kernel.update_params(samples)
        return predictive_mean_var_scalar(self.X, self.Y, self.kernel,self.jitter, X_test)
   
    def log_marginal_likelihood(self):
        """
        Compute the (negative) log marginal likelihood of the current model.

        Returns
        -------
        float
            The negative log marginal likelihood value.
        """
        return self.likelihood(self.X, self.Y, self.jitter, self.kernel)

    def sample_posterior(self, key: Key, X: Array, n_samples: Int) -> Array:
        """
        Sample from the posterior distribution at new input locations.

        Parameters
        ----------
        key : Key
            Random key for sampling.
        X : Array
            Input locations where to sample the posterior.
        n_samples : int
            Number of posterior samples to draw.

        Returns
        -------
        Array
            Posterior samples of shape (n_samples, n_test_samples, n_outputs).
        """
        return NotImplementedError
    
    def sample_prior(self, key: Key, X: Array, n_samples: Int) -> Array:
            """
            Sample from the prior distribution at new input locations.

            Parameters
            ----------
            key : Key
                Random key for sampling.
            X : Array
                Input locations where to sample the prior.
            n_samples : int
                Number of prior samples to draw.

            Returns
            -------
            Array
                Prior samples of shape (n_samples, n_test_samples, n_outputs).
            """
            return NotImplementedError

    def save(self, path: str):
        """
        Save the model parameters to a file.

        Parameters
        ----------
        path : str
            Path where the model parameters should be saved.
        """
        return NotImplementedError


# ------------------------------------------------------------------------------
# Project: IFE_Surrogate
# Authors: Tobias Leitgeb, Julian Tischler
# CD Lab 2025
# ------------------------------------------------------------------------------
from dataclasses import dataclass
from jaxtyping import Array, Key
from typing import Callable, Dict, List
from flax import struct
from .kernel import Kernel


#TODO
"""
There are problems with the naming scheme when u stack sum kernel and product kernel.
"""
@struct.dataclass
class KernelOperator(Kernel):
    r"""
    Represents a kernel that multiplies the outputs of two given kernels:
    
    .. math::

        k(x, x'|\theta_1, \theta_2) = k_1(x, x'|\theta_1) + k_2(x, x'|\theta_2)


    Attributes:
        kernel_1 (Kernel): The first kernel.
        kernel_2 (Kernel): The second kernel.
    """
    kernel_1: Kernel = struct.field(pytree_node=False)
    kernel_2: Kernel = struct.field(pytree_node=False)
    #priors: Dict[str, Callable] = struct.field(pytree_node=False)
    #TODO fix this somehow
    def __post_init__(self):
        #priors = {"k1": self.kernel_1.priors, "k2":  self.kernel_2.priors}
        #self.priors = priors
        pass

    def get_priors(self):
        """
        Get the kernel priors.

        Returns:
            A dictionary containing the kernel priors.
        """
        temp1 = self.kernel_1.get_priors()
        temp11 = {"k1__"+k : v for k, v in temp1.items()}

        temp2 = self.kernel_2.get_priors()
        temp22 = {"k2__"+k : v for k, v in temp2.items()}
        return temp11 | temp22

    
    def get_params(self) -> Dict[str, Array]:
        """
        Get the kernel parameters.

        Returns:
            A dictionary containing the kernel parameters.
        """
        temp1 = self.kernel_1.get_params()
        temp11 = {"k1__"+k : v for k, v in temp1.items()}

        temp2 = self.kernel_2.get_params()
        temp22 = {"k2__"+k : v for k, v in temp2.items()}
        return temp11 | temp22


    def sample_hyperparameters(self, key: Key) -> Dict[str, Array]:
        """
        Sample new parameters from the prior distribution.


        Args:
            key: JAX random key.

        Returns: 
            A dictionary containing sampled parameters.
        """
        
        # key1, key2 = jr.split(key, 2)

        params = self.get_params()
        prefixes = set([])
        for p in params.keys():
            prefixes.add(p.rsplit("__", 1)[0] + "__")

        kernels = [self.get_kernel(k) for k in prefixes]

        raw = {prefix : kernel.sample_hyperparameters(key) for prefix, kernel in zip(prefixes, kernels)}
        ret = {}
        for prefix, sample_item in raw.items():
            for name, sample in sample_item.items():
                ret[prefix + name] = sample

        return ret




    def get_kernel(self, path: str):
        if path.startswith("k1__"):
            if isinstance(self.kernel_1, KernelOperator):
                return self.kernel_1.get_kernel(path.split("__", 1)[1])
            else:
                return self.kernel_1
        elif path.startswith("k2__"):
            if isinstance(self.kernel_2, KernelOperator):
                return self.kernel_2.get_kernel(path.split("__", 1)[1])
            else:
                return self.kernel_2
        else:
            print("Warning: Invalid kernel call ", path, "Needs to start with 'k1__' or 'k2__'")
        


    
    def update_params(self, params: Dict) -> None:
        """
        Update the kernel parameters of both kernels.

        Args:
            params: A dictionary containing the new parameters.
        """
        
        for k, v in params.items():
            ksplit = k.split("__", 1)
            if len(ksplit) == 2:
                kernel_nr, param = ksplit
                if kernel_nr == "k1":
                    self.kernel_1.update_params({param: v})
                elif kernel_nr == "k2":
                    self.kernel_2.update_params({param: v})
            else:
                print("")







        

@struct.dataclass
class ProductKernel(KernelOperator):
    r"""
    Represents a kernel that multiplies the outputs of two given kernels:

    .. math::   

        k(x, x'|\theta_1, \theta_2) = k_1(x, x'|\theta_1) + k_2(x, x'|\theta_2)


    Attributes:
        kernel_1 (Kernel): The first kernel.
        kernel_2 (Kernel): The second kernel.
    """
    kernel_1: Kernel = struct.field(pytree_node=False)
    kernel_2: Kernel = struct.field(pytree_node=False)

    def evaluate(self, x1: Array, x2: Array) -> Array:
        return_1 = self.kernel_1(x1, x2)
        return_2 = self.kernel_2(x1, x2)
        return return_1 * return_2
    



@struct.dataclass
class SumKernel(KernelOperator):
    r"""
    Represents a kernel that multiplies the outputs of two given kernels:

    .. math::   

        k(x, x'|\theta_1, \theta_2) = k_1(x, x'|\theta_1) + k_2(x, x'|\theta_2)


    Attributes:
        kernel_1 (Kernel): The first kernel.
        kernel_2 (Kernel): The second kernel.
    """
    kernel_1: Kernel = struct.field(pytree_node=False)
    kernel_2: Kernel = struct.field(pytree_node=False)

    def evaluate(self, x1: Array, x2: Array) -> Array:
        return_1 = self.kernel_1(x1, x2)
        return_2 = self.kernel_2(x1, x2)
        return return_1 + return_2


@struct.dataclass
class SeparableKernel(KernelOperator):
    """
    A separable kernel that factors into a product of a kernel over x
    and a kernel over w:

    .. math::

        k((x, w), (x', w')) = k_x(x, x') \cdot k_w(w, w')

    Attributes:
        kx: Kernel acting on the x inputs.
        kw: Kernel acting on the w inputs.
    """
    #kx: Kernel = struct.field(pytree_node=False)
    #kw: Kernel = struct.field(pytree_node=False)


    def evaluate(self, inputs1, inputs2):
        """
        inputs1: tuple (X, W)
        inputs2: tuple (X, W)
        X: shape (N_x, D_x)
        W: shape (N_w, D_w)
        """
        X1, W1 = inputs1
        X2, W2 = inputs2
        Kx = self.kernel_1(X1, X2)
        Kw = self.kernel_2(W1, W2)
        return Kx, Kw 
    
    # def evaluate(self, x, w, x2, w2):
    #     return self.kernel_1(x, x2), self.kernel_2(w, w2)

    # def get_priors(self):
    #     px = {"kx__" + k: v for k, v in self.kx.get_priors().items()}
    #     pw = {"kw__" + k: v for k, v in self.kw.get_priors().items()}
    #     return px | pw

    # def get_params(self) -> Dict[str, Array]:
    #     px = {"kx__" + k: v for k, v in self.kx.get_params().items()}
    #     pw = {"kw__" + k: v for k, v in self.kw.get_params().items()}
    #     return px | pw

    # def sample_hyperparameters(self, key: Key) -> Dict[str, Array]:
    #     import jax.random as jr
    #     key_x, key_w = jr.split(key)
    #     px = {"kx__" + k: v for k, v in self.kx.sample_hyperparameters(key_x).items()}
    #     pw = {"kw__" + k: v for k, v in self.kw.sample_hyperparameters(key_w).items()}
    #     return px | pw

    # def update_params(self, params: Dict[str, Array]) -> None:
    #     px = {k.split("__", 1)[1]: v for k, v in params.items() if k.startswith("kx__")}
    #     pw = {k.split("__", 1)[1]: v for k, v in params.items() if k.startswith("kw__")}
    #     if px:
    #         self.kx.update_params(px)
    #     if pw:
    #         self.kw.update_params(pw)



# ------------------------------------------------------------------------------
# Project: IFE_Surrogate
# Authors: Tobias Leitgeb, Julian Tischler
# CD Lab 2025
# ------------------------------------------------------------------------------
from ife_surrogate.gp.models.gp_model import GPModel
from ife_surrogate.gp.train import train, train_scipy, train_swarm
from ife_surrogate.gp.likelihood import nmll_wideband, nlml_scalar
from ife_surrogate.gp.inference import predictive_mean_var_wideband, predictive_mean_var_scalar
from ife_surrogate.gp.kernels import Kernel
from ife_surrogate.utils.decorators import experimental, deprecated

from jaxtyping import Key, Array, Int, Bool
from typing import Tuple, Dict
from functools import partial
from jax import vmap
import typing
import jax.numpy as jnp
import numpyro
import optax
import numpyro.distributions as dist
from numpyro.infer import MCMC, NUTS
import warnings


TypeKernel = typing.TypeVar("Kernel", bound=Kernel)


class WidebandGP(GPModel):
    r"""
    Wideband Gaussian Process (GP) model for multi-output regression tasks.

    This GP handles multiple correlated outputs across a frequency domain, where the 
    outputs can be thought of as different frequency points.

    .. math::

        f \sim \mathcal{GP}(\mathbf{0}, \sigma_i^2 \otimes k(x, x'))

    Parameters
    ----------
    X : Array
        Training input data of shape (n_samples, n_features).
    Y : Array
        Training output data of shape (n_samples, n_outputs), where each output dimension
        typically corresponds to a different frequency.
    kernel : Kernel
        Instance of a kernel class defining the covariance structure.
    frequency : Array, optional
        Array of frequency values corresponding to the output dimensions. If provided, can
        be used to structure the multi-output modeling more explicitly.
    """

    def __init__(self, X: Array, Y: Array, kernel: Kernel, frequency: Array = None):
        # if X is not None and not X.shape[0] == Y.shape[0]:
        #     warnings.warn(f"X and Y should have the same number of samples. Is '{X.shape[0]}' != '{Y.shape[0]}'.", category=UserWarning)
        # if Y is not None and not Y.shape[1] > 1:
        #     warnings.warn(f"Y should be a matrix with shape (N, p) N: #samples, p: #outputs, disregard this warning if you need an empty model. Is '{X.shape[1]}' != '{Y.shape[1]}'.", category=UserWarning)
        super().__init__(kernel, X, Y)
        self.likelihood = nmll_wideband
        self.sigma_sq = Y.var(axis=0, ddof=1) if Y is not None else None
        self.jitter = 1e-6
        self.frequency = frequency

    @deprecated
    def train(
        self, 
        key: Key,
        optimizer: Dict = {"opt": optax.adam, "settings": {"learning_rate": 1e-2}, "tolerance": 1e-5, "patience": 20}, 
        sample_parameters: Bool = True,
        n_steps: Int = 1000, 
        n_restarts: int = 1, 
        save_history: Bool = False, 
        verbose: Bool = False
    ) -> Tuple:
        r"""
        Optax training function:
        Optimize kernel parameters by minimizing the negative marginal log-likelihood using gradient-based optimizers.

        Parameters
        ----------
        key : Key
            Random key for parameter initialization and restarts.
        optimizer : Dict, optional
            Dictionary containing the optimizer class (`"opt"`) and its settings (`"settings"`).
        sample_parameters : Bool
            Initial parameters for optimisation are sampled from the priors. If False then the current value of the hyperparameters is used to start the optimization.
        n_steps : int, optional
            Number of optimization steps for each restart.
        n_restarts : int, optional
            Number of independent optimization restarts with different initializations.
        save_history : bool, optional
            If True, saves the full optimization trajectory (parameter history).
        verbose : bool, optional
            If True, prints optimization progress.

        Returns
        -------
        Tuple
            Optimized parameters and, if requested, parameter history.
        """
        
        self.optimized_parameters, self.parameter_history = train(
            self, key, optimizer, sample_parameters, n_steps, n_restarts, save_history, verbose
        )
        self.kernel.update_params(self.optimized_parameters["params"])

    @deprecated
    def train_scipy(self, 
        key: Key, 
        n_restarts: int = 1, 
        optimizer: Dict = {"method": "L-BFGS-B"},
        save_history: Bool = False, 
        verbose: Bool = True
    ) -> None:
        r"""
        Scipy training function:
        Optimize kernel parameters using a scipy-based optimizer (e.g., L-BFGS-B).

        Parameters
        ----------
        key : Key
            Random key for parameter initialization and restarts.
        n_restarts : int, optional
            Number of independent optimization restarts with different initializations.
        optimizer : Dict, optional
            Dictionary specifying the optimizer method and its settings (e.g., `"method": "L-BFGS-B"`).
            (As defined in scipy: https://docs.scipy.org/doc/scipy/tutorial/optimize.html)
        save_history : bool, optional
            If True, saves the full optimization trajectory.
        verbose : bool, optional
            If True, prints optimization progress.

        Returns
        -------
        None
            The optimized hyperparameters are updated in the kernel.
        """
        self.optimized_parameters, self.parameter_history = train_scipy(
            self, key, n_restarts, optimizer, save_history, verbose
        )
        self.kernel.update_params(self.optimized_parameters["params"])
    
    @deprecated
    def train_swarm(self, 
        key: Key, 
        bounds: Tuple[Array, Array] = (),
        n_restarts: int = 1,
        optimizer: Dict = {"c1": 0.5, "c2": 0.3, "w": 0.9},
        n_particles: int = 20,
        n_iterations: int = 100,
        save_history: Bool = False, 
        verbose: Bool = True
    ) -> None:
        r"""
        Pyswarms training function:
        Optimize kernel parameters using PSO

        Parameters
        ----------
        key : Key
            Random key for parameter initialization and restarts.
        n_restarts : int, optional
            Number of independent optimization restarts with different initializations.
        optimizer : Dict, optional
            Dictionary specifying the optimization settings c1, c2, w as defined in pyswarms
                c1: particles own best position
                c2: global best position
                w: balance between c1, c2
                k: number of neighboring particles to consult
                p: distance metric (1: manhattan, 2: euclidean)
        n_particles : Int, optional
            Number of particles in the swarm.
        save_history : bool, optional
            If True, saves the full optimization trajectory.
        verbose : bool, optional
            If True, prints optimization progress.

        Returns
        -------
        None
            The optimized hyperparameters are updated in the kernel.
        """
        self.optimized_parameters, self.parameter_history = train_swarm(
            self, key, bounds, n_restarts=n_restarts, optimizer=optimizer, n_particles=n_particles, n_iterations=n_iterations,
        )
        self.kernel.update_params(self.optimized_parameters["params"])
    
    @deprecated
    def predict_old(self, X_test: Array) -> Tuple[Array, Array]:
        r"""
        Predict the mean and variance for the given test inputs.

        .. math::

            p(\mathbf{f}_{*,p} \mid \mathbf{X}, \mathbf{y}_p, \mathbf{X}_*)
            = \mathcal{N}\big(
                \mathbf{f}_* \mid \boldsymbol{\mu}_p, \boldsymbol{\Sigma}_p
            \big)

            \boldsymbol{\mu}_p
            = \mathbf{K}_*^{\mathsf{T}} \mathbf{K}_x^{-1} \mathbf{y}_p

            \boldsymbol{\Sigma}_p
            = \sigma_p^2 \big(
                \mathbf{K}_{**}
                - \mathbf{K}_*^{\mathsf{T}} \mathbf{K}_x^{-1} \mathbf{K}_*
            \big)

        Parameters
        ----------
        X_test : Array
            Test input data of shape (n_test_samples, n_features).

        Returns
        -------
        Tuple[Array, Array]
            Mean predictions and variances for each output dimension.
            Shapes: (n_test_samples, n_outputs)
        """
        pred_wideband = partial(
            predictive_mean_var_wideband,
            self.X,
            kernel=self.kernel,
            jitter=self.jitter,
            X_test=X_test,
        )
        sigma_sq = jnp.var(self.Y, axis=0, ddof=1)
        mean, var = vmap(pred_wideband, in_axes=(1, 0))(self.Y, sigma_sq)
        return mean.T, var.T
    
    
    def predict(self, X_test: Array) -> Tuple[Array, Array]:
        r"""
        Predict the mean and variance for the given test inputs.

        .. math::

            p(\mathbf{f}_{*,p} \mid \mathbf{X}, \mathbf{y}_p, \mathbf{X}_*) 
            = \mathcal{N}(\mathbf{f}_* \mid \boldsymbol{\mu}_p, \boldsymbol{\Sigma}_p)

        .. math::

            \boldsymbol{\mu}_p = \mathbf{K}_*^\top \mathbf{K}_x^{-1} \mathbf{y}_p

        .. math::

            \boldsymbol{\Sigma}_p 
            = \sigma_p^2 \left( \mathbf{K}_{**} - \mathbf{K}_*^\top \mathbf{K}_x^{-1} \mathbf{K}_* \right)

        Parameters
        ----------
        X_test : Array
            Test input data of shape ``(n_test_samples, n_features)``.

        Returns
        -------
        Tuple[Array, Array]
            Mean predictions and variances for each output dimension.
            Shapes: ``(n_test_samples, n_outputs)``.
        """
        pred_wideband = partial(predictive_mean_var_wideband, self.X, kernel=self.kernel, jitter=self.jitter, X_test=X_test)
        sigma_sq = jnp.var(self.Y, axis=0, ddof=1)
        mean, var = vmap(pred_wideband, in_axes=(1,0))(self.Y, sigma_sq)
        return mean.T, var.T

   
    def log_marginal_likelihood(self):
        r"""
        Compute the (negative) log marginal likelihood of the wideband model.

        .. math::

             -\log p(\mathbf{Y} \mid \mathbf{X}, \boldsymbol{\theta}, \boldsymbol{\sigma}^2)
            =
             -\sum_{p=1}^P \log\big(
                p(\mathbf{y}_p \mid \mathbf{X}, \boldsymbol{\theta}, \sigma_p^2)
            \big)

        Returns
        -------
        float
            The negative log marginal likelihood value.
        """
        return nmll_wideband(self.X, self.Y, self.sigma_sq, self.jitter, self.kernel)
    
    #@jit
    def log_likelihood_scalar(self):
        r"""
        Compute the (negative) log marginal likelihood assuming scalar outputs.

        .. math::
        
           \mathcal{L}_p =  p(\mathbf{y}_p|\mathbf{X}, \mathbf{\theta}, \sigma^2_p)

        Returns
        -------
        Array
            An array containing the value of the log likelihood at every task. If summed over it 
            is the same as the complete log likelihood.
        """
        return vmap(nlml_scalar, in_axes=(None, 1, None, None))(self.X, self.Y, self.jitter, self.kernel)

    def sample_posterior(self, key: Key, X: Array, n_samples: Int) -> Array:
        r"""
        Sample from the posterior distribution at new input locations.

        Parameters
        ----------
        key : Key
            Random key for sampling.
        X : Array
            Input locations where to sample the posterior.
        n_samples : int
            Number of posterior samples to draw.

        Returns
        -------
        Array
            Posterior samples of shape (n_samples, n_test_samples, n_outputs).
        """
        return NotImplementedError
    
    def sample_prior(self, key: Key, X: Array, n_samples: Int) -> Array:
        r"""
        Sample from the prior distribution at new input locations.

        Parameters
        ----------
        key : Key
            Random key for sampling.
        X : Array
            Input locations where to sample the prior.
        n_samples : int
            Number of prior samples to draw.

        Returns
        -------
        Array
            Prior samples of shape (n_samples, n_test_samples, n_outputs).
        """
        return NotImplementedError


@experimental
class WidebandGPBaysian(GPModel):
    r"""
    Bayesian Wideband Gaussian Process model using MCMC sampling.

    This model leverages full Bayesian inference over the kernel parameters
    using Hamiltonian Monte Carlo (HMC) sampling with the NUTS algorithm.

    Parameters
    ----------
    kernel : Kernel
        Kernel object specifying the covariance structure.
    X : Array
        Training input data of shape (n_samples, n_features).
    Y : Array
        Training output data of shape (n_samples, n_outputs).
    """

    def __init__(self, X, Y, kernel):
        super().__init__(kernel, X, Y)
        self.likelihood = nmll_wideband
        self.sigma_sq = Y.var(axis=0)
        self.jitter = 1e-6
    
    def model_forward(self,):
        r"""
        Defines the probabilistic model for Bayesian inference.

        Samples kernel parameters from their priors and defines the joint likelihood
        over all output frequencies under a multivariate normal distribution.

        Notes
        -----
        - Assumes that the data variance per output is known and fixed.
        - Kernel parameters (like "power" and "lengthscale") are sampled from priors.
        """
        n,p,d=self.X.shape[0], self.Y.shape[1], self.X.shape[1]
        parameters = self.kernel.get_params() #dictionary of all hyperparamers
        hyperparameters = parameters.keys() #list of names of hyperparameters
        prior_functions = self.kernel.get_priors() #dictionary of all prior functions

        for hyperparam_name in hyperparameters:
            n_features = parameters[hyperparam_name].shape #extracts the dimension of the hperparam. e.g ard kernel
            sampled_hyperparam = numpyro.sample(
                hyperparam_name, prior_functions[hyperparam_name].expand(n_features)
            )
            #new sampled hyperparam is put into the dictionary
            parameters[hyperparam_name] = sampled_hyperparam.copy()
            self.kernel.update_params(parameters)#kernel is updated

        
        K = self.kernel(self.X, self.X) + jnp.eye(n) * self.jitter
        
        normal = dist.MultivariateNormal(jnp.zeros(n), K)
        #here again we make the assumption that we dont need to use the variance as a hyperparameter,
        #but just use the data variance
        data_variances = jnp.var(self.Y, axis=0)

        with numpyro.plate("frequency_positions", p): #plate 
            #here I use a little trick so we do not have to calculate the normal distribution for each task
            # N(0, sigma^2) = N(0, 1) / sigma
            obs = self.Y/jnp.sqrt(data_variances)
            for i in range(p):
                numpyro.sample(
                    f"Y{i}", 
                    normal, 
                    obs=obs[:, i], 
                )

    def train(self, key: Key, num_samples:Int = 100, num_warmup:Int = 100):
        r"""
        Run MCMC to sample from the posterior distribution of the kernel parameters.

        Parameters
        ----------
        key : Key
            Random key for MCMC sampling.
        num_samples : int, optional
            Number of MCMC samples to collect after warm-up.
        num_warmup : int, optional
            Number of warm-up (burn-in) steps before sampling.

        Returns
        -------
        None
            Updates the model's `samples` attribute with posterior samples.
        """
        nuts_kernel = NUTS(self.model_forward)
        mcmc = MCMC(nuts_kernel, num_warmup=num_warmup, num_samples=num_samples)
        mcmc.run(key)
        mcmc.print_summary()
        samples = mcmc.get_samples()
        self.samples = samples


    def predict(self, X_test: Array):
        r"""
        Predict mean and variance at test points using posterior samples.

        For each posterior sample of the kernel parameters, predictions are made
        and aggregated.

        Parameters
        ----------
        X_test : Array
            Test input data of shape (n_test_samples, n_features).

        Returns
        -------
        Tuple[Array, Array]
            - Mean predictions for each MCMC sample: (n_samples, n_test_samples, n_outputs).
            - Variances for each MCMC sample: (n_samples, n_test_samples, n_outputs).
        """
        pred_one_sample = partial(predictive_mean_var_wideband, self.X, jitter=self.jitter, X_test=X_test)
        #calculates the prediction over the whole frequency axis
        vmap_over_frequency = vmap(pred_one_sample, in_axes=(1,0, None), out_axes=(1,1))

        def wrapper(params):
            self.kernel.update_params(params)
            return vmap_over_frequency(self.Y, self.sigma_sq, self.kernel)
        #calculates the predictions for every generated sample in self.samples
        y_pred_samples, y_var_samples = vmap(wrapper)(self.samples)

        return y_pred_samples, y_var_samples

    def sample_posterior(self, key: Key, X: Array, n_samples: Int) -> Array:
        r"""
        Sample from the posterior distribution at new input locations.

        Parameters
        ----------
        key : Key
            Random key for sampling.
        X : Array
            Input locations where to sample the posterior.
        n_samples : int
            Number of posterior samples to draw.

        Returns
        -------
        Array
            Posterior samples of shape (n_samples, n_test_samples, n_outputs).
        """
        return NotImplementedError
    
    def sample_prior(self, key: Key, X: Array, n_samples: Int) -> Array:
        r"""
        Sample from the prior distribution at new input locations.

        Parameters
        ----------
        key : Key
            Random key for sampling.
        X : Array
            Input locations where to sample the prior.
        n_samples : int
            Number of prior samples to draw.

        Returns
        -------
        Array
            Prior samples of shape (n_samples, n_test_samples, n_outputs).
        """
        return NotImplementedError

    def save(self, path: str):
        r"""
        Save the model parameters to a file.

        Parameters
        ----------
        path : str
            Path where the model parameters should be saved.
        """
        return NotImplementedError




# ------------------------------------------------------------------------------
# Project: IFE_Surrogate
# Authors: Tobias Leitgeb, Julian Tischler
# CD Lab 2025
# ------------------------------------------------------------------------------
from .data_loader import train_test_split
from .metrics import nrms, rmse
from .plotting.visualizer import Visualizer
from .fsv import apply_fsv
from .preprocessing import zero_mean_unit_var_axis0, MinMaxScaler, inverse_standardize_axis0
#from .sensitivity_analysis import gsa_sobol_analysis, gsa_fast_analysis

__all__ = [
    "train_test_split",
    "nrms",
    "rmse",
    "Visualizer",
    "apply_fsv",
    "zero_mean_unit_var_axis0",
    "MinMaxScaler",
]


# ------------------------------------------------------------------------------
# Project: IFE_Surrogate
# Authors: Tobias Leitgeb, Julian Tischler
# CD Lab 2025
# ------------------------------------------------------------------------------
from dataclasses import dataclass
from jaxtyping import Array
from typing import Callable, Dict
import jax.numpy as jnp
from .kernel import Kernel
from flax import struct

@struct.dataclass
class RBF(Kernel):
    r"""
    Radial Basis Function (RBF) Kernel:

    .. math::
        
        k(\mathbf{x}_1, \mathbf{x}_2) = \exp\left( -\frac{\lVert \mathbf{x}_1 - \mathbf{x}_2 \rVert^2}{2\ell^2} \right)
    

    Args:
        lengthscale (Array): Lengthscale controlling the smoothness of the kernel.
        priors (Dict[str, Callable], optional): Priors for hyperparameters.
    """
    lengthscale: Array# = struct.field(default_factory=lambda: jnp.array([]))
    # _priors: Dict[str, Callable] = struct.field(pytree_node=False, default_factory=dict)
    # _param_bounds: Dict = struct.field(pytree_node=False, default_factory=dict)

    def evaluate(self, x1: Array, x2: Array) -> Array:
        """
        Evaluate the Radial Basis Function (RBF) kernel between x1 and x2.

        Args:
            x1 (Array): Input array of shape (n_samples_1, n_features).
            x2 (Array): Input array of shape (n_samples_2, n_features).

        Returns:
            Array: Kernel matrix of shape (n_samples_1, n_samples_2).
        """
        sq_dist = jnp.sum(jnp.abs(x1[:, None] - x2)**2 * jnp.array(self.lengthscale)**-2, axis=-1)
        return jnp.array(1.0 * jnp.exp(-0.5 * sq_dist))






# ------------------------------------------------------------------------------
# Project: IFE_Surrogate
# Authors: Tobias Leitgeb, Julian Tischler
# CD Lab 2025
# ------------------------------------------------------------------------------
from .gp import kernels, models, trainers, inference, likelihood
from .utils import data_loader, metrics, plotting, preprocessing

__all__ = [
    "kernels", 
    "models",
    "trainers",

    "inference", 
    "likelihood",

    "data_loader",
    "parameter_model",
    "metrics",
    "plotting",
    "preprocessing"
]


# ------------------------------------------------------------------------------
# Project: IFE_Surrogate
# Authors: Tobias Leitgeb, Julian Tischler
# CD Lab 2025
# ------------------------------------------------------------------------------
from ife_surrogate.gp.models.gp_model import GPModel
from ife_surrogate.gp.train import train, train_scipy, train_swarm
from ife_surrogate.gp.likelihood import nlml_scalar, nmll_kron
from ife_surrogate.gp.inference import  predictive_mean_var_kron
from ife_surrogate.gp.kernels import Kernel
from ife_surrogate.utils.decorators import experimental, deprecated

from jaxtyping import Key, Array, Int, Bool
from typing import Tuple, Dict
import optax
from numpyro.infer import MCMC, NUTS
import warnings
from jax import vmap


@experimental
class SeparableMultiOutputGP(GPModel):
    r"""
    Separable Multi-Output Gaussian Process.

    .. math::
        :label: eq:nmll-kron

        \begin{aligned}
            \text{Model:} \quad
            & \operatorname{vec}(Y) \sim \mathcal{N}\big(0, K_w \otimes K_x \big) \\[1em]
            \text{NMLL:} \quad
            & \mathcal{L}(Y; K_x, K_w) \\
            & = \tfrac{1}{2} \operatorname{vec}(Y)^{\mathsf{T}} \big( K_w^{-1} \otimes K_x^{-1} \big) \operatorname{vec}(Y)
              + \tfrac{1}{2}\log \big| K_w \otimes K_x \big|
              + \tfrac{NP}{2}\log(2\pi) \\[0.5em]
            & = \tfrac{1}{2} \operatorname{Tr}\big( K_x^{-1} Y K_w^{-1} Y^{\mathsf{T}} \big)
              + \tfrac{1}{2} \Big( P \log |K_x| + N \log |K_w| \Big)
              + \tfrac{NP}{2}\log(2\pi)
        \end{aligned}

    Parameters
    ----------
    X : Array
        Training input data of shape (n_samples, n_features).
    Y : Array
        Training output data of shape (n_samples, n_outputs).
    kernel : Kernel
        Instance of a kernel class defining the covariance structure.
    frequency : Array, optional
        Array of frequency values corresponding to the output dimensions.
    """
    def __init__(self, X: Array, Y: Array, kernel: Kernel, frequency: Array = None):
        if X is not None and not X[0].shape[0] == Y[0].shape[0]:
            warnings.warn("X and Y should have the same number of samples.", category=UserWarning)
        if Y is not None and not Y.shape[1] > 1:
            warnings.warn("Y should be a matrix with shape (N, p) N: #samples, p: #outputs, disregard this warning if you need an empty model.", category=UserWarning)
        super().__init__(kernel, X, Y)
        self.likelihood = nmll_kron
        #self.sigma_sq = Y.var(axis=0, ddof=1) if Y is not None else None
        self.jitter = 1e-6
        self.frequency = frequency
        

    def train(
        self, 
        key: Key,
        optimizer: Dict = {"opt": optax.adam, "settings": {"learning_rate": 1e-2}, "tolerance": 1e-5, "patience": 20}, 
        sample_parameters: Bool = True,
        n_steps: Int = 1000, 
        n_restarts: int = 1, 
        save_history: Bool = False, 
        verbose: Bool = False
    ) -> Tuple:
        r"""
        Optax training function:
        Optimize kernel parameters by minimizing the negative marginal log-likelihood using gradient-based optimizers.

        Parameters
        ----------
        key : Key
            Random key for parameter initialization and restarts.
        optimizer : Dict, optional
            Dictionary containing the optimizer class (`"opt"`) and its settings (`"settings"`).
        sample_parameters : Bool
            Initial parameters for optimisation are sampled from the priors. If False then the current value of the hyperparameters is used to start the optimization.
        n_steps : int, optional
            Number of optimization steps for each restart.
        n_restarts : int, optional
            Number of independent optimization restarts with different initializations.
        save_history : bool, optional
            If True, saves the full optimization trajectory (parameter history).
        verbose : bool, optional
            If True, prints optimization progress.

        Returns
        -------
        Tuple
            Optimized parameters and, if requested, parameter history.
        """
        
        self.optimized_parameters, self.parameter_history = train(
            self, key, optimizer, sample_parameters, n_steps, n_restarts, save_history, verbose
        )
        self.kernel.update_params(self.optimized_parameters["params"])


    def train_scipy(self, 
        key: Key, 
        n_restarts: int = 1, 
        optimizer: Dict = {"method": "L-BFGS-B"},
        save_history: Bool = False, 
        verbose: Bool = True
    ) -> None:
        r"""
        Scipy training function:
        Optimize kernel parameters using a scipy-based optimizer (e.g., L-BFGS-B).

        Parameters
        ----------
        key : Key
            Random key for parameter initialization and restarts.
        n_restarts : int, optional
            Number of independent optimization restarts with different initializations.
        optimizer : Dict, optional
            Dictionary specifying the optimizer method and its settings (e.g., `"method": "L-BFGS-B"`).
            (As defined in scipy: https://docs.scipy.org/doc/scipy/tutorial/optimize.html)
        save_history : bool, optional
            If True, saves the full optimization trajectory.
        verbose : bool, optional
            If True, prints optimization progress.

        Returns
        -------
        None
            The optimized hyperparameters are updated in the kernel.
        """
        self.optimized_parameters, self.parameter_history = train_scipy(
            self, key, n_restarts, optimizer, save_history, verbose
        )
        self.kernel.update_params(self.optimized_parameters["params"])
    

    def train_swarm(self, 
        key: Key, 
        bounds: Tuple[Array, Array],
        n_restarts: int = 1,
        optimizer: Dict = {"c1": 0.5, "c2": 0.3, "w": 0.9},
        n_particles: int = 20,
        n_iterations: int = 100,
        save_history: Bool = False, 
        verbose: Bool = True
    ) -> None:
        r"""
        Pyswarms training function:
        Optimize kernel parameters using PSO

        Parameters
        ----------
        key : Key
            Random key for parameter initialization and restarts.
        n_restarts : int, optional
            Number of independent optimization restarts with different initializations.
        optimizer : Dict, optional
            Dictionary specifying the optimization settings c1, c2, w as defined in pyswarms
                c1: particles own best position
                c2: global best position
                w: balance between c1, c2
                k: number of neighboring particles to consult
                p: distance metric (1: manhattan, 2: euclidean)
        n_particles : Int, optional
            Number of particles in the swarm.
        save_history : bool, optional
            If True, saves the full optimization trajectory.
        verbose : bool, optional
            If True, prints optimization progress.

        Returns
        -------
        None
            The optimized hyperparameters are updated in the kernel.
        """
        self.optimized_parameters, self.parameter_history = train_swarm(
            self, key, bounds, n_restarts=n_restarts, optimizer=optimizer, n_particles=n_particles, n_iterations=n_iterations,
        )
        self.kernel.update_params(self.optimized_parameters["params"])
    
    @deprecated
    def predict_old(self, XW_test: Tuple[Array, Array]) -> Tuple[Array, Array]:
        r"""
        Predict the mean and variance for the given test inputs.

        .. math::

            p(\mathbf{f}_{*,p} \mid \mathbf{X}, \mathbf{y}_p, \mathbf{X}_*)
            =
            \mathcal{N}\big(
                \mathbf{f}_* \mid \boldsymbol{\mu}_p, \boldsymbol{\Sigma}_p
            \big)

            \boldsymbol{\mu}_p
            =
            \mathbf{K}_*^{\mathsf{T}} \mathbf{K}_x^{-1} \mathbf{y}_p

            \boldsymbol{\Sigma}_p
            =
            \sigma_p^2 \big(
                \mathbf{K}_{**}
                -
                \mathbf{K}_*^{\mathsf{T}} \mathbf{K}_x^{-1} \mathbf{K}_*
            \big)

        Parameters
        ----------
        XW_test : Tuple[Array, Array]
            Tuple containing the test inputs ``X_test`` and corresponding frequencies
            (or auxiliary inputs) ``f_test``.

        Returns
        -------
        Tuple[Array, Array]
            Mean predictions and variances for each output dimension.
            Shapes: (n_test_samples, n_outputs)
        """
   
        
        y_pred, var_pred = predictive_mean_var_kron(self.X, self.Y, self.kernel, self.jitter, XW_test)
        return y_pred, var_pred
    
    def predict(self, XW_test: Tuple[Array, Array]) -> Tuple[Array, Array]:
        r"""
        Predict the mean and variance for the given test inputs.

        .. math::

            p(\mathbf{f_{*,p}}|\mathbf{X},\mathbf{y_p},\mathbf{X_*})
            = \mathcal{N}(\mathbf{f_*}|\mathbf{\mu_{p}},\mathbf{\Sigma_p})

        .. math::

            \mathbf{\mu_p}
            = \mathbf{K_*}^{T}\mathbf{K_x}^{-1} \mathbf{y_p}

        .. math::

            \mathbf{\Sigma_p}
            = \sigma_p^2 (\mathbf{K_{**}} - \mathbf{K_*}^{T}\mathbf{K_x}^{-1}\mathbf{K_*})

        Parameters
        ----------
        XW_test : Tuple(Array, Array)
            A tuple containing ``X_test`` and ``f_test``.

        Returns
        -------
        Tuple[Array, Array]
            Mean predictions and variances for each output dimension.
            Shapes: ``(n_test_samples, n_outputs)``.
        """
        
        y_pred, var_pred = predictive_mean_var_kron(
            self.X, self.Y, self.kernel, self.jitter, XW_test
        )
        return y_pred, var_pred
   
    def log_marginal_likelihood(self):
        r"""
        Compute the (negative) log marginal likelihood of the wideband model.

        .. math::

             -\log p(\mathbf{Y} \mid \mathbf{X}, \boldsymbol{\theta}, \boldsymbol{\sigma}^2)
            =
             -\sum_{p=1}^P \log\big(
                p(\mathbf{y}_p \mid \mathbf{X}, \boldsymbol{\theta}, \sigma_p^2)
            \big)

        Returns
        -------
        float
            The negative log marginal likelihood value.
        """
        return nmll_kron(self.X, self.Y, self.jitter, self.kernel)
    
    #@jit
    def log_likelihood_scalar(self):
        r"""
        Compute the (negative) log marginal likelihood assuming scalar outputs.

        .. math::
        
           \mathcal{L}_p =  p(\mathbf{y}_p|\mathbf{X}, \mathbf{\theta}, \sigma^2_p)

        Returns
        -------
        Array
            An array containing the value of the log likelihood at every task. If summed over it 
            is the same as the complete log likelihood.
        """
        return vmap(nlml_scalar, in_axes=(None, 1, None, None))(self.X, self.Y, self.jitter, self.kernel)

    def sample_posterior(self, key: Key, X: Array, n_samples: Int) -> Array:
        r"""
        Sample from the posterior distribution at new input locations.

        Parameters
        ----------
        key : Key
            Random key for sampling.
        X : Array
            Input locations where to sample the posterior.
        n_samples : int
            Number of posterior samples to draw.

        Returns
        -------
        Array
            Posterior samples of shape (n_samples, n_test_samples, n_outputs).
        """
        return NotImplementedError
    
    def sample_prior(self, key: Key, X: Array, n_samples: Int) -> Array:
            """
            Sample from the prior distribution at new input locations.

            Parameters
            ----------
            key : Key
                Random key for sampling.
            X : Array
                Input locations where to sample the prior.
            n_samples : int
                Number of prior samples to draw.

            Returns
            -------
            Array
                Prior samples of shape (n_samples, n_test_samples, n_outputs).
            """
            return NotImplementedError

    def save(self, path: str):
        """
        Save the model parameters to a file.

        Parameters
        ----------
        path : str
            Path where the model parameters should be saved.
        """
        return NotImplementedError




# ------------------------------------------------------------------------------
# Project: IFE_Surrogate
# Authors: Tobias Leitgeb, Julian Tischler
# CD Lab 2025
#
# Base class for Gaussian Process Regressor Models
#
# ------------------------------------------------------------------------------
from ..kernels import Kernel

from jaxtyping import Array, Key
import typing
from abc import ABC, abstractmethod
import joblib


TypeKernel = typing.TypeVar("Kernel", bound=Kernel)



class GPModel(ABC):
    r""" 
        Abstract base class of a GP, to be used as skeleton for actual GP implementations.
    """

    def __init__(self, kernel: TypeKernel, X: Array, Y: Array):
        self.kernel = kernel
        self.X = X
        self.Y = Y

    ## Must be implemented in actual model implementation, every model should have a predict method
    @abstractmethod
    def predict():
        return

    def get_attributes(self):
        return self.__dict__

    #TODO 
    def save(self, filename="IfeSurrModel.pkl"):
        with open(filename, "wb") as output:
            joblib.dump(self, filename)
    
    @staticmethod
    def load(filename):
        return joblib.load(filename)


import matplotlib.pyplot as plt
from jaxtyping import Array
import seaborn as sns
from matplotlib.backends.backend_pdf import PdfPages
import numpy as np
import matplotlib as mpl
from collections.abc import Iterable
from ...utils.metrics import nrms, rmse



class Visualizer:
    """
    Container class for useful plotting functions.
    Can also export figures to a PDF-File.
    """

    def save_to_page(self,
                     figures, 
                     file_name: str = "output.pdf"):
        """Function to export matplotlib figures to PDF

        Args:
            figures (iterable[matplotlib.pyplot.figure]): List of matplotlib figures, each 
                                                          figure adds a new page to the output PDF
                                                          Also allows single figure.
            file_name (str, optional): Path and name of the PDF
                                       Defaults to "output.pdf".
        """
        try:
            iter(figures)
            
            with PdfPages(file_name) as pdf:
                for figure in figures:
                    pdf.savefig(figure)
                    plt.close()
        except TypeError:
            with PdfPages(file_name) as pdf:
                pdf.savefig(figures)
                plt.close()

    def plot_bounded_error(self, x, y, error,
                           function_color="blue",
                           error_color="lightblue", 
                           highlight_threshold=0, 
                           highlight_color="red"):
        """Default bounded error plot.

        Args:
            x (_type_): _description_
            y (_type_): _description_
            error (_type_): _description_

        Returns:
            matplotlib.pyplot.figure: figure can be adjusted afterwards.
        """
        fig, ax = plt.subplots()

        ax.plot(x, y, color=function_color, label="Function", linewidth=1)
        sns.set_theme(style="white")

        ax.fill_between(x, y - error, y + error, label="Error Bounds", color=error_color, alpha=0.75)

        if highlight_threshold != 0:
            ax.fill_between(x,
                            y - error,
                            y + error,
                            label="Error Bounds", color=highlight_color, where=np.array([True if i > highlight_threshold else False for i in error]), alpha=0.75)
        
        ax.set_title("Title", fontsize=14)
        ax.set_xlabel("X-Axis", fontsize=12)
        ax.set_ylabel("Y-Axis", fontsize=12)
        ax.legend(fontsize=12)
        ax.grid(True, linestyle="--", alpha=0.5)
        return fig
    

    def plot_sns(self):
        fig, ax = plt.subplots()
        sns.set_theme(style="darkgrid")
        fmri = sns.load_dataset("fmri")

        # Plot the responses for different events and regions
        sns.lineplot(x="timepoint", y="signal",
             hue="region", style="event",
             data=fmri, ax=ax)
        return fig


    def plot_testset(
            self,
            frequency: Array, 
            test_y: Array, 
            prediction: Array, 
            var_pred: Array = None, 
            grid: tuple = (4, 4),
            figsize: tuple = (15, 15),
            log_scale: bool = True,
            error: str = "nrms"
        )-> plt.figure:
        """
        Plot predicted vs true curves for a test set.

        Each subplot shows the true and predicted frequency response for one test sample.
        If predictive variance is given, 95% confidence intervals (±2 std) are shown.

        Args:
            frequency (Array): 1D array of frequency values of shape (p,).
            test_y (Array): True output values of shape (n_test, p).
            prediction (Array): Predicted output values of shape (n_test, p).
            var_pred (Array, optional): Predictive variances of shape (n_test, p).
                                        If provided, uncertainty bands are shown.
            grid (tuple): Grid size as (rows, cols) to arrange the subplots.
            figsize (tuple): Size of the overall figure.
            log_scale (bool): If True, x-axis is displayed in logarithmic scale.
            error (str): Type of error metric to show in the title of each plot.
                        Currently only supports "rmse" and "nrms.

        Returns:
            matplotlib.figure.Figure: The generated matplotlib figure.
        """
        fig, ax  = plt.subplots(grid[0], grid[1], figsize=figsize)
        
        if grid[0] == 1 and grid[1] == 1:
            ax = np.array([ax])
        else:
            ax = ax.flatten()

        for i in range(grid[0]*grid[1]):
            ax[i].plot(frequency, test_y[i, :], color='black')
            ax[i].plot(frequency, prediction[i, :], color='blue')
            if var_pred is not None:
                pred_std = np.sqrt(var_pred[i, :])
                ax[i].fill_between(
                    frequency, 
                    prediction[i, :] - 2*pred_std, 
                    prediction[i, :] + 2*pred_std, 
                    color='blue', alpha=0.2
                )
        
            ax[i].set_ylabel('dB')
            ax[i].legend(['True', 'Predicted'])
            #ax[i].grid()
            if log_scale:
                ax[i].set_xscale('log')
            if error == "rmse":
                test_error = rmse(test_y[i, :], prediction[i, :]).item() * 100
                ax[i].set_title(f'Index: {i}, {error}: {test_error:.3f} %')
            elif error == "nrms":
                test_error = nrms(test_y[i, :], prediction[i, :]).item() * 100
                ax[i].set_title(f'Index: {i}, {error}: {test_error:.3f} %')
            else:
                pass
            #add axis only for the last row
            if i >= grid[0]*(grid[1]-1):
                ax[i].set_xlabel('Frequency (Hz)')
            plt.tight_layout()
        return fig


    @staticmethod
    def quickplot(data_dict: dict, fig_height=2, fig_width=7):
        for key in data_dict.keys():
            ## Assume the axis with most entries is x-axis
            # print("length of ", len(data_dict[key].shape))
            data = np.array(data_dict[key])
            if data.ndim == 1:
                data = data.reshape(-1, 1)
            nr_plots = min(data.shape)
            xaxis = np.argmax(data.shape)
            yaxis = np.argmin(data.shape)
            # print("selectes axis: ", xaxis)

            # print(slices)
            proceed = input(f"Plot {nr_plots} plots for key '{key}'? (y/n)")
            if proceed == "y":
                fig, ax = plt.subplots(nr_plots, 1)
                if not isinstance(ax, Iterable):
                    ax = [ax]

                fig.set_figheight(fig_height*nr_plots)
                fig.set_figwidth(fig_width)

                # indices = [i for i in range(dt.shape[xaxis])]
                moved = np.moveaxis(data, yaxis, 0)
                for i in range(nr_plots):
                    
                    ax[i].scatter([i for i in range(len(moved[i]))], moved[i], color="k", marker="x", s=1)
                    ax[i].title.set_text(f"{key}: {i}")
                    # ax[i].scatter([i for i in range(dt.shape[xaxis])], dt[tuple(slices)])
                plt.show()
            else:
                print("Aborted.")
                






# ------------------------------------------------------------------------------
# Project: IFE_Surrogate
# Authors: Tobias Leitgeb, Julian Tischler
# CD Lab 2025
# ------------------------------------------------------------------------------
"""
Contains all internal kernels for the GPR.
New Kernels can be created by inheriting from the base kernel 'Kernel'

Example:
    >>> from ife_surrogate.gp.kernels import Kernel
    >>> from flax import struct
    >>> 
    >>> @struct.dataclass
    >>> class NewKernel(Kernel):
    >>>     def evaluate(self, x1: Array, x2: Array) -> Array:
    >>>         # some kernel definition
    >>>         return []
"""
from .kernel import Kernel
from .rbf import RBF
from .kriging_kernel import Kriging
from .rational_quadratic import RQ
from .matern import Matern12, Matern32, Matern52
from .compound_kernels import SumKernel, ProductKernel, SeparableKernel
from .scale import Scale
from .noise import DiagNoise
from .fixed_kernel import FixedFreqKernel


__all__ = [
    'Kernel',
    'RBF', 
    'Kriging', 
    'RQ',
    'Matern12',
    'Matern32',
    'Matern52',
    'SumKernel',
    'ProductKernel',
    'Scale',
    'DiagNoise',
    'SeparableKernel',
    'FixedFreqKernel'
]


import jax.numpy as jnp

def zero_mean_unit_var_axis0(Y, eps=1e-8):
    """
    Standardize Y along axis 0 (samples) so each feature/output 
    has zero mean and unit variance.

    Parameters
    ----------
    Y : array, shape (N, P)
        Input data (N samples, P outputs).
    eps : float
        Small constant to avoid division by zero.

    Returns
    -------
    Y_scaled : array, shape (N, P)
        Standardized data.
    mean : array, shape (1, P)
        Mean of each output across samples.
    std : array, shape (1, P)
        Std of each output across samples.
    """
    mean = jnp.mean(Y, axis=0, keepdims=True)
    std = jnp.std(Y, axis=0, keepdims=True)
    Y_scaled = (Y - mean) / (std + eps)
    return Y_scaled, mean, std

def inverse_standardize_axis0(Y_scaled, mean, std, eps=1e-8):
    """
    Backtransform standardized data to original scale.
    """
    return Y_scaled * (std + eps) + mean


class MinMaxScaler:
    def __init__(self, feature_range=(0.0, 1.0), eps=1e-8):
        self.feature_range = feature_range
        self.eps = eps
        self.data_min_ = None
        self.data_max_ = None
        self.scale_ = None
        self.min_ = None

    def fit(self, Y, axis=None):
        """Compute min and max for scaling."""
        Y_min = jnp.min(Y, axis=axis, keepdims=True)
        Y_max = jnp.max(Y, axis=axis, keepdims=True)
        
        self.data_min_ = Y_min
        self.data_max_ = Y_max
        data_range = Y_max - Y_min
        
        fr_min, fr_max = self.feature_range
        self.scale_ = (fr_max - fr_min) / (data_range + self.eps)
        self.min_ = fr_min - Y_min * self.scale_
        return self

    def transform(self, Y):
        """Scale data to the feature range."""
        return Y * self.scale_ + self.min_

    def inverse_transform(self, Y_scaled):
        """Revert back to the original scale."""
        return (Y_scaled - self.min_) / self.scale_


from flax import struct
import jax.numpy as jnp
from typing import Dict
from jaxtyping import Array
from .kernel import Kernel
from jax.scipy.interpolate import RegularGridInterpolator

## TODO Underscore does not work anymore
## DUPLICATE?? 
@struct.dataclass
class FixedFreqKernel(Kernel):
    """
    Kernel with a learnable spatial part Kx and a *fixed* frequency correlation matrix Cw.
    Kw = diag(alpha) @ Cw @ diag(alpha).
    """
    sigma: Array 
    C_w: Array = struct.field(pytree_node=False, default_factory=dict)                             # fixed correlation matrix (P, P)
    # _priors: Dict[str, callable] = struct.field(pytree_node=False, default_factory=dict)
    # _param_bounds: Dict = struct.field(pytree_node=False, default_factory=dict)

    # full evaluate (returns both Kx and Cw)
    def evaluate(self, W1, W2):

        # frequency part (fixed, just slice the known C_w)
        idx1 = jnp.searchsorted(W1.ravel(), W1.ravel())
        idx2 = jnp.searchsorted(W2.ravel(), W2.ravel())
        Cw = self.C_w[jnp.ix_(idx1, idx2)]

        return Cw
    
@struct.dataclass
class FixedFreqKernel(Kernel):
    """
    Kernel with a learnable spatial part Kx and a *fixed* frequency correlation matrix Cw.
    Kw = diag(alpha) @ Cw @ diag(alpha).

    The frequency correlation is interpolated from a reference correlation matrix C_w
    defined on W_ref.
    """
    sigma: Array
    _C_w: Array = struct.field(pytree_node=False)     # fixed correlation matrix (P,P)
    # _W_ref: Array = struct.field(pytree_node=False)   # reference frequency grid
    # _priors: Dict[str, callable] = struct.field(pytree_node=False, default_factory=dict)
    # _param_bounds: Dict = struct.field(pytree_node=False, default_factory=dict)

    def evaluate(self, W1, W2):
        """
        Interpolates Cw(W1, W2) from reference C_w.
        """
        # convert to numpy because RegularGridInterpolator is scipy, not jax
        W1 = jnp.asarray(W1).ravel()
        W2 = jnp.asarray(W2).ravel()
        W_ref = jnp.asarray(self._W_ref)
        C_w = jnp.asarray(self._C_w) 
        C_w += jnp.eye(C_w.shape[0])*self.sigma

        # Build interpolator *locally*, not as a field
        interp = RegularGridInterpolator(
            (W_ref, W_ref),
            C_w,
            bounds_error=False,
            fill_value=None
        )

        # Interpolate
        points = jnp.array([[(w1, w2) for w2 in W2] for w1 in W1]).reshape(-1, 2)
        Cw_eval = interp(points).reshape(len(W1), len(W2))

        # return as jax array for further computation
        return jnp.asarray(Cw_eval) 





# ------------------------------------------------------------------------------
# Project: IFE_Surrogate
# Authors: Tobias Leitgeb, Julian Tischler
# CD Lab 2025
# ------------------------------------------------------------------------------
import warnings
import functools

class ExperimentalWarning(UserWarning):
    """Warning for experimental API usage."""
    pass


def experimental(obj):
    """
    Decorator to mark functions/classes as experimental.

    Emits a warning on first use.
    """
    message = f"{obj.__name__} is experimental and may change or be removed in future versions."

    if isinstance(obj, type):
        ## Decorating a class
        orig_init = obj.__init__

        @functools.wraps(orig_init)
        def new_init(self, *args, **kwargs):
            warnings.warn(message, ExperimentalWarning, stacklevel=2)
            return orig_init(self, *args, **kwargs)

        obj.__init__ = new_init
        obj.__experimental__ = True
        return obj

    else:
        ## Decorating a function
        @functools.wraps(obj)
        def wrapper(*args, **kwargs):
            warnings.warn(message, ExperimentalWarning, stacklevel=2)
            return obj(*args, **kwargs)

        wrapper.__experimental__ = True
        return wrapper



class DeprecatedWarning(UserWarning):
    """Warning for deprecated API usage."""
    pass


def deprecated(deprecated_in=None, removed_in=None, use_instead=None):
    """
    Decorator to mark a function/class as deprecated.

    Parameters
    ----------
    deprecated_in : str, optional
        Version in which it was deprecated.
    removed_in : str, optional
        Version in which it will be removed.
    use_instead : str, optional
        Name of the recommended replacement.
    """
    def decorator(obj):
        msg = f"{obj.__name__} is deprecated"
        if deprecated_in:
            msg += f" since version {deprecated_in}"
        if removed_in:
            msg += f" and will be removed in version {removed_in}"
        if use_instead:
            msg += f". Use {use_instead} instead."
        msg += "."

        if isinstance(obj, type):
            ## classes
            init = obj.__init__

            @functools.wraps(init)
            def new_init(self, *args, **kwargs):
                warnings.warn(msg, DeprecatedWarning, stacklevel=2)
                return init(self, *args, **kwargs)

            obj.__init__ = new_init
            obj.__deprecated__ = True
            return obj

        else:
            ## functions
            @functools.wraps(obj)
            def wrapper(*args, **kwargs):
                warnings.warn(msg, DeprecatedWarning, stacklevel=2)
                return obj(*args, **kwargs)

            wrapper.__deprecated__ = True
            return wrapper

    return decorator



# ------------------------------------------------------------------------------
# Project: IFE_Surrogate
# Authors: Tobias Leitgeb, Julian Tischler
# CD Lab 2025
# ------------------------------------------------------------------------------
from typing import Dict, Callable
from dataclasses import dataclass
from jaxtyping import Array
import jax.numpy as jnp
from flax import struct
from .kernel import Kernel


@struct.dataclass
class Matern12(Kernel):
    r"""
    Matérn 1/2 kernel (Exponential kernel).

    Computes the covariance between x1 and x2 as:

    .. math::
        k(\mathbf{x}_1, \mathbf{x}_2) = \exp\left(-\frac{\lVert \mathbf{x}_1 - \mathbf{x}_2 \rVert}{\ell}\right)

    Args:
        lengthscale (Array): Lengthscale controlling the smoothness.
        priors (Dict[str, Callable], optional): Priors for hyperparameters.
    """
    lengthscale: Array
    # _priors: Dict[str, Callable] = struct.field(pytree_node=False)
    # _param_bounds: Dict = struct.field(pytree_node=False, default_factory=dict)

    def evaluate(self, x1: Array, x2: Array) -> Array:
        """
        Evaluate the Matérn 1/2 kernel.

        Args:
            x1 (Array): Input array of shape (n_samples_1, n_features).
            x2 (Array): Input array of shape (n_samples_2, n_features).

        Returns:
            Array: Kernel matrix of shape (n_samples_1, n_samples_2).
        """
        dist = jnp.sum(jnp.abs(x1[:, None] - x2) / self.lengthscale, axis=-1)
        return jnp.exp(-dist)


@struct.dataclass 
class Matern32(Kernel):
    r"""
    Matérn 3/2 kernel.

    Computes the covariance between x1 and x2 as:

    .. math::
        k(\mathbf{x}_1, \mathbf{x}_2) = \left(1 + \frac{\sqrt{3} \lVert \mathbf{x}_1 - \mathbf{x}_2 \rVert}{\ell} \right)
        \exp\left(-\frac{\sqrt{3} \lVert \mathbf{x}_1 - \mathbf{x}_2 \rVert}{\ell} \right)

    Args:
        lengthscale (Array): Lengthscale controlling the smoothness.
        priors (Dict[str, Callable], optional): Priors for hyperparameters.
    """
    lengthscale: Array 
    # _priors: Dict[str, Callable] = struct.field(pytree_node=False)
    # _param_bounds: Dict = struct.field(pytree_node=False, default_factory=dict)

    def evaluate(self, x1: Array, x2: Array) -> Array:
        """
        Evaluate the Matérn 3/2 kernel.

        Args:
            x1 (Array): Input array of shape (n_samples_1, n_features).
            x2 (Array): Input array of shape (n_samples_2, n_features).

        Returns:
            Array: Kernel matrix of shape (n_samples_1, n_samples_2).
        """
        dist = jnp.sum(jnp.abs(x1[:, None] - x2) / self.lengthscale, axis=-1)
        return (1.0 + jnp.sqrt(3.0) * dist) * jnp.exp(-jnp.sqrt(3.0) * dist)


@struct.dataclass
class Matern52(Kernel):
    r"""
    Matérn 5/2 kernel.

    Computes the covariance between x1 and x2 as:

    .. math::
        k(\mathbf{x}_1, \mathbf{x}_2) = \left(1 + \frac{\sqrt{5} \lVert \mathbf{x}_1 - \mathbf{x}_2 \rVert}{\ell} + \frac{5 \lVert \mathbf{x}_1 - \mathbf{x}_2 \rVert^2}{3 \ell^2} \right)
        \exp\left(-\frac{\sqrt{5} \lVert \mathbf{x}_1 - \mathbf{x}_2 \rVert}{\ell} \right)


    Produces GP sample paths that are twice differentiable.

    Args:
        lengthscale (Array): Lengthscale controlling the smoothness.
        priors (Dict[str, Callable], optional): Priors for hyperparameters.
    """
    lengthscale: Array 
    # _priors: Dict[str, Callable] = struct.field(pytree_node=False)
    # _param_bounds: Dict = struct.field(pytree_node=False, default_factory=dict)

    def evaluate(self, x1: Array, x2: Array) -> Array:
        """
        Evaluate the Matérn 5/2 kernel.

        Args:
            x1 (Array): Input array of shape (n_samples_1, n_features).
            x2 (Array): Input array of shape (n_samples_2, n_features).

        Returns:
            Array: Kernel matrix of shape (n_samples_1, n_samples_2).
        """
        scaled_diff = (x1[:, None] - x2) / self.lengthscale
        dist = jnp.sqrt(jnp.sum(scaled_diff**2, axis=-1))
        sqrt5_dist = jnp.sqrt(5.0) * dist
        return (1.0 + sqrt5_dist + (5.0 / 3.0) * dist**2) * jnp.exp(-sqrt5_dist)



# ------------------------------------------------------------------------------
# Project: IFE_Surrogate
# Authors: Tobias Leitgeb, Julian Tischler
# CD Lab 2025
# ------------------------------------------------------------------------------
from jaxtyping import Array, Key
import jax.random as jr
from typing import Tuple
import jax.numpy as jnp

def train_test_split(
            X: Array,
            Y: Array,
            f: Array = jnp.array([]),
            dense: bool = False,
            *,
            split: Tuple[float, float, float] = (0.8, 0.2, 0.0),
            key: Key,
        )-> Tuple[Tuple[Array, Array], Tuple[Array, Array], Tuple[Array, Array]]:
    """
    Splits the dataset into training, testing, and validation sets.

    Parameters
    ----------
    X : Array
        Input feature array.
    Y : Array
        Target value array.
    f : Array
        Additional feature array.
    dense : bool, optional
        If True, the input features are repeated and concatenated with ``f``.
        Default is False.
    split : Tuple[float, float, float], optional
        Ratios for splitting the data into (train, test, validation).
        Default is ``(0.8, 0.2, 0.0)``.
    key : Key
        Random key for shuffling the data.

    Returns
    -------
    Tuple[Tuple[Array, Array], Tuple[Array, Array], Tuple[Array, Array]]
        A tuple of three pairs:
        ``(X_train, Y_train)``, ``(X_test, Y_test)``, and ``(X_val, Y_val)``.
    """

    n = X.shape[0]
    n_train = int(n * split[0])
    n_test = int(n * split[1])
   

    #shuffle data
    key, subkey = jr.split(key)
    idx = jr.permutation(subkey, n)
    X = X[idx, :]
    Y = Y[idx, :]

    #split data
    X_train, Y_train = jnp.array(X[:n_train, :]), jnp.array(Y[:n_train, :])
    X_test, Y_test = jnp.array(X[n_train:n_train + n_test, :]), jnp.array(Y[n_train:n_train + n_test, :])
    X_val, Y_val = jnp.array(X[n_train + n_test:, :]), jnp.array(Y[n_train + n_test:, :])

    if dense:
        X_train = jnp.repeat(X_train, len(f), axis=0)
        X_train = jnp.concatenate([X_train, jnp.tile(f, len(X_train)//len(f))[:,None]], axis=1)
        X_test = jnp.repeat(X_test, len(f), axis=0)
        X_test = jnp.concatenate([X_test, jnp.tile(f, len(X_test)//len(f))[:,None]], axis=1)
        X_val = jnp.repeat(X_val, len(f), axis=0)
        X_val = jnp.concatenate([X_val, jnp.tile(f, len(X_val)//len(f))[:,None]], axis=1)
        return (X_train, Y_train), (X_test, Y_test), (X_val, Y_val)
    else:
        return (X_train, Y_train), (X_test, Y_test), (X_val, Y_val)



# ------------------------------------------------------------------------------
# Project: IFE_Surrogate
# Authors: Tobias Leitgeb, Julian Tischler
# CD Lab 2025
# ------------------------------------------------------------------------------
import numpy as np
from numpy.fft import fft, ifft
from typing import List

def find_point(array):
    """
    Find the index in 'array' where cumulative sum reaches 40% of total sum.
    """
    total_sum = np.sum(np.abs(array))
    target = 0.4 * total_sum
    cum = 0
    for idx, val in enumerate(np.abs(array)):
        cum += val
        if cum >= target:
            return idx
    return len(array) - 1


def _apply_fsv_single(y1, y2):
    """Apply FSV on one pair of signals."""
    N = len(y1)

    # FFT
    F1 = fft(y1)
    F2 = fft(y2)

    # determine split frequency (ignore first 4 bins)
    idx1 = find_point(F1[4:]) + 4
    idx2 = find_point(F2[4:]) + 4
    split = min(idx1, idx2)

    # build low/high filters with linear ramp over ±3 bins
    lo_filter = np.zeros(N)
    hi_filter = np.ones(N)
    start = max(split - 3, 0)
    end = min(split + 3, N - 1)
    ramp = np.linspace(1, 0, end - start + 1)
    lo_filter[start:end + 1] = ramp
    hi_filter[start:end + 1] = 1 - ramp

    # apply filters and inverse FFT
    lo1 = np.real(ifft(F1 * lo_filter))
    lo2 = np.real(ifft(F2 * lo_filter))
    hi1 = np.real(ifft(F1 * hi_filter))
    hi2 = np.real(ifft(F2 * hi_filter))

    # ADM
    alpha = np.abs(lo1 - lo2)
    beta = (1/N) * np.sum(np.abs(lo1) + np.abs(lo2))
    X = np.abs(y1 - y2)
    delta = (1/N) * np.sum(np.abs(y1) + np.abs(y2))
    ADMi = (alpha / beta) + (X / delta) * np.exp(X / delta)
    ADMi[ADMi > 50] = 0
    ADM = np.mean(ADMi)

    # derivatives
    def deriv(arr, order=1):
        if order == 1:
            d = np.zeros_like(arr)
            d[1:-1] = arr[2:] - arr[:-2]
            d[0] = arr[1] - arr[0]
            d[-1] = arr[-1] - arr[-2]
            return d
        elif order == 2:
            return deriv(deriv(arr, 1), 1)
        else:
            raise ValueError("Only orders 1 and 2 supported")

    dlo1 = deriv(lo1, 1); dlo2 = deriv(lo2, 1)
    dhi1 = deriv(hi1, 1); dhi2 = deriv(hi2, 1)
    d2hi1 = deriv(hi1, 2); d2hi2 = deriv(hi2, 2)

    # FDM
    FDM1 = np.abs(dlo1 - dlo2) / ((2/N) * np.sum(np.abs(dlo1) + np.abs(dlo2)))
    FDM2 = np.abs(dhi1 - dhi2) / ((6/N) * np.sum(np.abs(dhi1) + np.abs(dhi2)))
    FDM3 = np.abs(d2hi1 - d2hi2) / ((7.2/N) * np.sum(np.abs(d2hi1) + np.abs(d2hi2)))
    FDMi = 2 * (FDM1 + FDM2 + FDM3)
    FDM = np.mean(FDMi)

    # GDM
    GDMi = np.sqrt(ADMi**2 + FDMi**2)
    GDM = np.mean(GDMi)

    return ADM, FDM, GDM


def apply_fsv(Y_test: np.ndarray, Y_pred: np.ndarray, score_strings: bool = False)-> np.ndarray:
    """
    Batch FSV: apply across multiple test/pred pairs.
    Calculates the FSV parameters as described in [1]
    
    [1] A. P. Duffy, A. J. M. Martin, A. Orlandi, G. Antonini, T. M. Benson, and M. S. Woolfson,
    "Feature Selective Validation (FSV) for Validation of Computational Electromagnetics (CEM).
    Part I—The FSV Method," IEEE Trans. Electromagnetic Compatibility, vol. 48, no. 3,
    pp. 449-458, Aug. 2006, doi:10.1109/TEMC.2006.879358 

    Args:
        Y_test: array of shape (M, N) or list of arrays
        Y_pred: same shape as Y_test
        score_string: returns the labels instead of the numerical scores

    Returns:
        ADM_list, FDM_list, GDM_list: arrays of length M
    """
    Y_test = np.asarray(Y_test, dtype=float)
    Y_pred = np.asarray(Y_pred, dtype=float)
    if Y_test.ndim == 1:
        return _apply_fsv_single(Y_test, Y_pred)

    M = Y_test.shape[0]
    ADM_list = np.zeros(M)
    FDM_list = np.zeros(M)
    GDM_list = np.zeros(M)
    for i in range(M):
        ADM_list[i], FDM_list[i], GDM_list[i] = _apply_fsv_single(Y_test[i], Y_pred[i])

    if score_strings:
        return categorize_scores(ADM_list), categorize_scores(FDM_list), categorize_scores(GDM_list)
    else:
        return ADM_list, FDM_list, GDM_list


def categorize_scores(scores: np.ndarray, bins=None, labels=None) -> List[str]:
    """
    Map numeric scores to qualitative labels based on bins.

    Supports scalars or arrays; returns single label or list correspondingly.

    Args:
        scores: scalar or array-like
        bins: list of bin edges; default [0,0.1,0.2,0.4,0.8,1.6,inf]
        labels: list of labels; default ['Excellent','Very Good','Good','Fair','Poor','Very Poor']

    Returns:
        Single label (str) if input scalar, else list of labels.
    """
    if bins is None:
        bins = [0, 0.1, 0.2, 0.4, 0.8, 1.6, np.inf]
    if labels is None:
        labels = ['Excellent', 'Very Good', 'Good', 'Fair', 'Poor', 'Very Poor']

    arr = np.atleast_1d(scores)
    idx = np.digitize(arr, bins, right=False) - 1
    idx = np.clip(idx, 0, len(labels)-1)
    mapped = [labels[i] for i in idx]
    return mapped[0] if np.isscalar(scores) else mapped



# ------------------------------------------------------------------------------
# Project: IFE_Surrogate
# Authors: Tobias Leitgeb, Julian Tischler
# CD Lab 2025
# ------------------------------------------------------------------------------
from typing import Callable, Tuple
import jax.numpy as jnp
import jax.scipy as jsp
from jax import vmap
from jaxtyping import Array, Float

from ife_surrogate.utils.decorators import experimental


def nmll_wideband(
        X: Array,
        Y: Array,
        sigma_sq: Array,
        jitter: Float,
        kernel: Callable,
    ) -> Float:
    r"""
    Computes the negative marginal log-likelihood (NMLL) for a wideband (multi-output) Gaussian Process model:

    .. math::
        
        -\log p(\mathbf{Y} | \mathbf{X}, \mathbf{\theta}, \mathbf{\sigma}^2) = -\sum_{p=1}^P \log\left( p(\mathbf{y}_p|\mathbf{X}, \mathbf{\theta}, \sigma^2_p) \right)

    with:
    
    .. math::

        p(\mathbf{y}_p|\mathbf{X}, \mathbf{\theta}, \sigma^2_p) = 
        \frac{1}{2} \mathbf{y}_p^\intercal \mathbf{K}^{-1} \mathbf{y}_p 
        + \frac{1}{2} \log\left\lvert \mathbf{K} \right\rvert 
        + \frac{n}{2} \log(2\pi) 
        + \frac{N}{2} \log(\sigma_p^2)

    Args:
        X (Array): Training input data of shape (N, D).
        Y (Array): Training output data of shape (N, P), where P is the number of outputs (tasks).
        sigma_sq (Array): Per-output noise variance, shape (P,).
        jitter (float): Small positive constant added to the diagonal of the kernel matrix for numerical stability.
        kernel (Callable): Kernel function used to compute covariances between inputs.

    Returns:
        Float: The negative marginal log-likelihood value.
    
    Notes:
        This implementation assumes independent noise across outputs (diagonal noise covariance),
        and scales the likelihood calculation accordingly.
    """
    def calc_inner_loops(Y, L):
        r"""
        Calculate the inner loops of a function.

        Parameters:
        - Y_minus_mean: The difference between Y and the mean value.
        - L: The matrix L.

        Returns:
        - The result of the calculation.
            """
        return 0.5 * jnp.dot((Y).T, jsp.linalg.cho_solve((L, True), Y))
    
    inner_loop = vmap(calc_inner_loops, in_axes=(1, None))
    
    p_ = Y.shape[1]
    N = Y.shape[0]
    K = kernel(X, X)
    L = jsp.linalg.cho_factor(K + jitter*jnp.eye(N), lower=True)[0]
    
    nmll = (N*p_/2 * jnp.log(2*jnp.pi) 
            +  p_ * jnp.sum(jnp.log((jnp.diag(L)))) 
            + N/2 * jnp.sum(jnp.log(sigma_sq))) 
    # print(sigma_sq)
    nmll += jnp.sum(inner_loop(Y, L) * (sigma_sq)**-1) #there was a + jitter for some reason
    # nmll += jnp.sum(inner_loop(Y, L) * sigma_sq**-1)
    return nmll


def nlml_scalar(
        X: Array,
        Y: Array,
        jitter: Float,
        kernel: Callable,
    ) -> Float:
    r"""
    Computes the negative marginal log-likelihood (NMLL) for a scalar-output Gaussian Process model:

    .. math::

        p(\mathbf{y}|\mathbf{X}, \mathbf{\theta}) = 
        \frac{1}{2} \mathbf{y}^\intercal \mathbf{K}^{-1} \mathbf{y}
        + \frac{1}{2} \log\left\lvert \mathbf{K} \right\rvert 
        + \frac{N}{2} \log(2\pi) 
        

    Args:
        X (Array): Training input data of shape (N, D).
        Y (Array): Training output data of shape (N, 1) or (N,).
        sigma_sq (Array): Estimated noise variance, shape (1,) or scalar.
        jitter (float): Small positive constant added to the diagonal of the kernel matrix for numerical stability.
        kernel (Callable): Kernel function used to compute covariances between inputs.

    Returns:
        Float: The negative marginal log-likelihood value.
    
    Notes:
        This function is tailored for standard (single-output) Gaussian Process models,
        assuming a homoskedastic noise model.
    """
    N = X.shape[0]
    K = kernel(X, X)
    L = jsp.linalg.cho_factor(K + jitter*jnp.eye(N), lower=True)
    alpha = jsp.linalg.cho_solve(L, Y)
    nmll = (
        N/2 * jnp.log(2*jnp.pi) 
        +jnp.sum(jnp.log((jnp.diag(L[0])))) 
        +1/2*jnp.dot(Y.T, alpha)
    )
    return nmll.squeeze()


@experimental
def nmll_kron(
        X: Tuple[Array, Array],
        Y: Array,
        jitter: Float,
        kernel: Callable
    ) -> Float:
    r"""
    Computes the negative marginal log-likelihood (NMLL) for a wideband (multi-output)
    Gaussian Process model: with K_w = diag(alpha) C_w diag(alpha) with alpha = sqrt(sigma**2)

    .. math::

        -\log p(\mathbf{Y} | \mathbf{X}, \mathbf{\theta}, \mathbf{\sigma}^2)
        = -\sum_{p=1}^P \log\left( p(\mathbf{y}_p|\mathbf{X}, \mathbf{\theta}, \sigma^2_p) \right)

    with:

    .. math::

        \text{vec}(\mathbf{Y})^\top 
        \left( \mathbf{K}_t^{-1} \otimes \mathbf{K}_x^{-1} \right) 
        \text{vec}(\mathbf{Y})
        &= \mathrm{trace} \left( \mathbf{K}_x^{-1} \, \mathbf{Y} \, \mathbf{K}_t^{-1} \, \mathbf{Y}^\top \right) \\[6pt]
        &= \sum_{i=1}^N \sum_{j=1}^P 
        \left[ \mathbf{K}_x^{-1} \mathbf{Y} \right]_{ij} \;
        \left[ \mathbf{K}_t^{-1} (\mathbf{K}_x^{-1} \mathbf{Y})^\top \right]_{ji}

    Args:
        X (Array): Training input data of shape (N, D).
        Y (Array): Training output data of shape (N, P), where P is the number of outputs (tasks).
        jitter (float): Small positive constant added to the diagonal of the kernel matrix for numerical stability.
        kernel (Callable): Kernel function used to compute covariances between inputs.

    Returns:
        Float: The negative marginal log-likelihood value.

    Notes:
        This implementation assumes independent noise across outputs (diagonal noise covariance),
        and scales the likelihood calculation accordingly.
    """

    
    N, P = Y.shape
    X, W = X
    sigma_sq = jnp.var(Y, axis=0,ddof=1)    
    alpha_diag = jnp.diag(jnp.sqrt(sigma_sq))

    K_x, C_w = kernel.evaluate((X, W), (X, W))

    K_w = alpha_diag @ C_w @ alpha_diag
    L_x = jsp.linalg.cho_factor(K_x + jitter)
    L_w = jsp.linalg.cho_factor(K_w + jitter)#* jnp.diag(sigma_sq))

    log_det = P * 2*jnp.sum(jnp.log(jnp.diag(L_x[0]))) \
            + N * 2*jnp.sum(jnp.log(jnp.diag(L_w[0])))
    
    A = jsp.linalg.cho_solve(L_x, Y)      # (N,P)
    M = A.T @ Y                           # (P,P)
    quad_form = jnp.trace(jsp.linalg.cho_solve(L_w, M))


    nmll = 0.5 * (quad_form + log_det + N*P*jnp.log(2*jnp.pi))
    
    return nmll





# ------------------------------------------------------------------------------
# Project: IFE_Surrogate
# Authors: Tobias Leitgeb, Julian Tischler
# CD Lab 2025
# ------------------------------------------------------------------------------
from ife_surrogate.gp.trainers import Trainer
from ife_surrogate.gp.models import GPModel

from functools import partial
import typing
from typing import Callable, Dict, Tuple, Optional
import optax
from jax import value_and_grad, jit, random
from typing import Tuple, Dict, Callable
from jaxtyping import Key, Array, Float, Int, Bool
import jax.numpy as jnp
import jax.random as jr


TypeGPModel = typing.TypeVar("GPModel", bound=GPModel)


class OptaxTrainer(Trainer):
    r"""
    Trainer for Gaussian Process models using gradient-based optimization with Optax[1].

    This class leverages JAX and Optax to optimize model parameters using
    gradient descent, supporting multiple restarts and early stopping.

    Attributes:
        key (Key): JAX random key for parameter sampling and reproducibility.
        optimizer: An Optax optimizer instance (e.g., optax.adam).[2]
        number_iterations (int): Maximum number of iterations per training run.
        number_restarts (int): Number of independent training restarts.
        tolerance (float): Minimum improvement threshold to reset early stopping.
        patience (int): Number of iterations to wait without improvement before stopping.
        verbose (bool): Whether to print iteration progress.
        save_history (bool): Whether to return the full optimization history.
        sample_parameters (bool): Whether to sample kernel hyperparameters at each restart.

    Methods:
        train(model: GPModel) -> Tuple[Dict, Optional[Dict]]:
            Trains the given GP model and returns the best parameters and optional history.
        info():
            Prints the current trainer settings for easy reference.
    [1] "https://optax.readthedocs.io/en/latest/"
    [2] "https://optax.readthedocs.io/en/latest/api/optimizers.html"
    """
    
    def __init__(self, 
                 key: Key = jr.key(seed=0),
                 optimizer = optax.adam(1e-3),
                 number_iterations: int = 100,
                 number_restarts: int = 1,
                 tolerance: Float = 1e-2, 
                 patience: Float = 20, **kwargs):
        super().__init__(**kwargs)
        self.key = key
        self.optimizer = optimizer
        self.tolerance = tolerance
        self.patience = patience
        self.number_iterations = number_iterations
        self.number_restarts = number_restarts

    def info(self):
        """Prints current settings of the OptaxTrainer."""
        print("OptaxTrainer settings:")
        print(f"\tkey: {self.key}")
        print(f"\toptimizer: {self.optimizer}")
        print(f"\tnumber_iterations: {self.number_iterations}")
        print(f"\tnumber_restarts: {self.number_restarts}")
        print(f"\ttolerance: {self.tolerance}")
        print(f"\tpatience: {self.patience}")
        print(f"\tverbose: {self.verbose}")
        print(f"\tsave_history: {self.save_history}")
        print(f"\tsample_parameters: {self.sample_parameters}")

    def train(self, model: TypeGPModel) -> Tuple[Dict, Optional[Dict]]:
        self.model = model
        value_and_grad_fn = value_and_grad(self.loss_fn)

        @jit
        def step(params: Dict, opt_state: Dict):
            loss, grads = value_and_grad_fn(params)
            updates, opt_state = self.optimizer.update(grads, opt_state)
            params = optax.apply_updates(params, updates)
            return params, opt_state, loss

        def loop(params, opt_state):
            best_loss = jnp.inf
            no_improve = 0
            for it in range(self.number_iterations):
                params, opt_state, loss = step(params, opt_state)

                if loss + self.tolerance < best_loss:
                    best_loss = loss
                    no_improve = 0
                else:
                    no_improve += 1

                if self.verbose and it % 50 == 0:
                    print(f"Iter {it}: loss={loss:.5f}, best={best_loss:.5f}")

                if no_improve >= self.patience:
                    if self.verbose:
                        print(f"Early stop at iter {it}, best loss {best_loss:.5f}")
                    break
            return params, opt_state, best_loss

        results = {}
        for r in range(self.number_restarts):
            _, self.key = random.split(self.key)
            if self.sample_parameters:
                params = self.model.kernel.sample_hyperparameters(self.key)
            else:
                params = self.model.kernel.get_params()

            opt_state = self.optimizer.init(params)
            params, opt_state, loss = loop(params, opt_state)

            results[f"run_{r}"] = {"params": params, "loss": loss, "key": self.key}
            if self.verbose:
                print(f"Run {r} loss: {loss}")

        results = {k: v for k, v in results.items() if not jnp.isnan(v["loss"])}
        best_key = min(results, key=lambda x: results[x]["loss"])

        if self.verbose:
            print("Best run:", results[best_key]["loss"])

        if self.save_history:
            return results[best_key], results
        return results[best_key], results





# ------------------------------------------------------------------------------
# Project: IFE_Surrogate
# Authors: Tobias Leitgeb, Julian Tischler
# CD Lab 2025
# ------------------------------------------------------------------------------
import optax
from jax import value_and_grad, jit, random
from typing import Tuple, Dict, Callable
from functools import partial
from jaxtyping import Key, Array, Float, Int, Bool
import numpy as np
import jax.numpy as jnp
from scipy.optimize import minimize
from tqdm import tqdm
import pyswarms as ps
from jax import vmap
from jax.flatten_util import ravel_pytree
import logging

from ife_surrogate.utils.decorators import deprecated

logging.getLogger("pyswarms.single.global_best").setLevel(logging.WARNING)

## Old version for training, use *Trainer


@deprecated
def train(
        model: Callable, 
        key: Key, 
        optimizer: Dict,
        sample_parameters: Bool = True,
        number_iterations: Int =10000,
        number_restarts: int=1,
        save_history: Bool=False, 
        verbose: Bool=True,
    ) -> Tuple:
    """
    Trains a model using Optax optimizers with multiple random restarts.

    Args:
        model (Callable): The GP model containing kernel, likelihood, and data (X, Y, etc.).
        key (Key): JAX random key for initialization.
        optimizer (Dict): Dictionary with "opt" (Optax optimizer constructor) and "settings" (hyperparameters).
        sample_parameters(Bool). Flag if a sample of the priors is used as a starting point for the optimisation.
        number_iterations (int): Number of optimization steps per restart.
        number_restarts (int): Number of random restarts.
        save_history (bool): If True, returns all optimization runs' history.
        verbose (bool): If True, prints progress and loss values.

    Returns:
        Tuple: (best_run: Dict, history: Dict or None)
            - best_run: Dictionary containing "params", "loss", and "key" for the best performing restart.
            - history: If save_history is True, dictionary of all runs; otherwise None.
    """
    kernel = model.kernel
    likelihood = model.likelihood
    if "sigma_sq" in model.get_attributes().keys():
        nlml = partial(likelihood, model.X, model.Y, model.sigma_sq, model.jitter)
    else:
        nlml = partial(likelihood, model.X, model.Y, model.jitter)
    
    def loss_fn(params: Dict) -> float:
        kernel.update_params(params)
        return nlml(kernel)

    value_and_grad_fn = value_and_grad(loss_fn)

    opt = optimizer["opt"](**optimizer["settings"])
    tolerance = optimizer["tolerance"]
    patience = optimizer["patience"]

    @jit
    def step(params: Dict, opt_state: Dict) -> Tuple:
        loss, grads = value_and_grad_fn(params)
        updates, opt_state = opt.update(grads, opt_state)
        params = optax.apply_updates(params, updates)
        #print(params)
        return params, opt_state, loss
    
    def loop(params: Dict, opt_state: Dict) -> Tuple:
        best_loss = jnp.inf
        no_improve = 0
        for it in range(number_iterations):
            params, opt_state, loss = step(params, opt_state)

            # Early stopping check
            if loss + tolerance < best_loss:  # improvement found
                best_loss = loss
                no_improve = 0
            else:
                no_improve += 1

            if verbose and it % 50 == 0:
                print(f"Iter {it}: loss={loss:.5f}, best={best_loss:.5f}")

            if no_improve >= patience:
                #if verbose:
                print(f"Stopping early at iter {it}, best loss {best_loss:.5f}")
                break

        return params, opt_state, best_loss
    
    #make several restarts with different initializations
    dict_params = {}
    for _ in (range(number_restarts)):
        _, key = random.split(key) 
        if sample_parameters:
            params = kernel.sample_hyperparameters(key)
        else:
            params = kernel.get_params()
        opt_state = opt.init(params)
        params, opt_state, loss = loop(params, opt_state)
        dict_optimization_run = {"params": params, "loss": loss, "key": key}
        dict_params["run_{}".format(_)] = dict_optimization_run
        if verbose:
            print(f"Loss: {loss}")
    #remove nans
    dict_params = {k: v for k, v in dict_params.items() if not jnp.isnan(v['loss'])}
    best_run = min(dict_params, key=lambda x: dict_params[x]['loss'])
    if verbose:
        print("Best run: ", dict_params[best_run]['loss'])
    if save_history:
        print("All optimization runs saved.")
        return dict_params[best_run], dict_params
    return dict_params[best_run], None
    

@deprecated
def train_scipy( 
        model: Callable, 
        key: Key, 
        number_restarts: int=1, 
        opt_algorithm: Dict={"method":"L-BFGS-B"},
        save_history: Bool=False, 
        verbose: Bool=True
    ) -> Tuple:
    """
    Trains a model using Scipy's minimize function with multiple random restarts.

    Args:
        model (Callable): The GP model containing kernel, likelihood, and data (X, Y, etc.).
        key (Key): JAX random key for initialization.
        number_restarts (int): Number of random restarts.
        opt_algorithm (Dict): Dictionary specifying the optimization method for Scipy minimize.
        save_history (bool): If True, returns all optimization runs' history.
        verbose (bool): If True, prints progress and loss values.

    Returns:
        Tuple: (best_run: Dict, history: Dict)
            - best_run: Dictionary containing "params", "loss", and "key" for the best performing restart.
            - history: Dictionary of all optimization runs.
    """
    kernel = model.kernel
    likelihood = model.likelihood
    if "sigma_sq" in model.get_attributes().keys():
        nlml = partial(likelihood, model.X, model.Y, model.sigma_sq, model.jitter)
    else:
        nlml = partial(likelihood, model.X, model.Y, model.jitter)

    dict_params = {}
    # function for raveling and unraveling the hyperparameter dictionary of the kernel
    params_template = kernel.get_params()
    _, unravel_fn = ravel_pytree(params_template)

    def loss_fn(flat_params: Array) -> float:
        #params = array_to_dict(params)
        params = unravel_fn(flat_params)
        kernel.update_params(params)
        return nlml(kernel)

    for i in tqdm(range(number_restarts)):
        key, subkey = random.split(key)
        params = kernel.sample_hyperparameters(subkey)
        flat_params, _ = ravel_pytree(params)
     
        res = minimize(
            loss_fn, 
            flat_params,
            **opt_algorithm,
            )
        optimized_params = unravel_fn(res.x)
        dict_optimization_run = {"params": optimized_params, "loss": res.fun, "key": key}
        dict_params["run_{}".format(i)] = dict_optimization_run
        if verbose:
            print(f"Loss: {res.fun}")
    
    dict_params = {k: v for k, v in dict_params.items() if not jnp.isnan(v['loss'])}
    best_run = min(dict_params, key=lambda x: dict_params[x]['loss'])
    print("Best run: ", best_run, ":", dict_params[best_run]['loss'])
    return dict_params[best_run], dict_params


@deprecated
def train_swarm(
        model: Callable,
        key: Key,
        bounds: Tuple[Array, Array] = (),
        optimizer: Dict = {"c1": 0.5, "c2": 0.3, "w": 0.9},
        n_restarts: Int = 1,
        n_particles: Int = 20,
        n_iterations: Int = 100,
        save_history: bool = False,
        verbose: bool = True
    ) -> Tuple:
    """
    Trains a model using PySwarms particle swarm optimizer with multiple random restarts.

    Args:
        model (Callable): The GP model containing kernel, likelihood, and data (X, Y, etc.).
        key (KeyArray): JAX random key for initialization.
        bounds (Tuple(Array, Array)) bounds for the optimization, Arrays must be of shape dimension 
        number_restarts (int): Number of random restarts.
        optimizer (Dict): Optimizer options for PySwarms (c1, c2, w).
        number_particles (int): Number of particles.
        number_iterations (int): Number of iterations.
        save_history (bool): If True, returns all optimization runs' history.
        verbose (bool): If True, prints progress and loss values.

    Returns:
        Tuple: (best_run: Dict, history: Dict)
            - best_run: Dictionary containing "params", "loss", and "key" of the best performing restart.
            - history: Dictionary of all optimization runs.
    """
    kernel = model.kernel
    likelihood = model.likelihood

    if "sigma_sq" in model.get_attributes().keys():
        nlml = partial(likelihood, model.X, model.Y, model.sigma_sq, model.jitter)
    else:
        nlml = partial(likelihood, model.X, model.Y, model.jitter)
    

    # function for raveling and unraveling the hyperparameter dictionary of the kernel
    dict_params = dict()
    params_template = kernel.get_params()
    raveled_template, unravel_fn = ravel_pytree(params_template)
    dimensions = len(raveled_template)
    
    
    def single_particle_loss(flat_params: Array) -> float:
        #params = array_to_dict(params)
        params = unravel_fn(flat_params)
        kernel.update_params(params)
        return nlml(kernel)
    loss_fn = jit(vmap(single_particle_loss))

    # for i in tqdm(range(n_restarts)):
    # key, subkey = random.split(key)

    param_keys = params_template.keys()
    priors_template = kernel.get_priors()

    if bounds == ():
        lower_bounds = []
        upper_bounds = []
        for prior_name in priors_template:
            if prior_name in param_keys:
                args = priors_template[prior_name].get_args()
                lower = args.get("low")
                upper = args.get("high")
                diff = abs(upper - lower)
                lower_bounds.extend([lower-diff] * len(params_template[prior_name]))
                upper_bounds.extend([upper+diff] * len(params_template[prior_name]))
        bounds = (lower_bounds, upper_bounds)

    # Particle Swarm Optimization via PySwarms
    optimizer_ = ps.single.GlobalBestPSO(
        n_particles=n_particles,
        dimensions=dimensions,
        options=optimizer,
        bounds=bounds
    )
    i=1
    best_cost, best_pos = optimizer_.optimize(loss_fn, iters=n_iterations, verbose=True)

    optimized_params= unravel_fn(best_pos)

    dict_optimization_run = {"params": optimized_params, "loss": best_cost, "key": key}
    dict_params[f"run_{i}"] = dict_optimization_run

    #if verbose:
    #    print(f"Run {i}: Loss = {best_cost}")

    # Filter out NaN runs if any
    dict_params = {k: v for k, v in dict_params.items() if not np.isnan(v['loss'])}

    # Find the best run
    best_run = min(dict_params, key=lambda x: dict_params[x]['loss'])
    print("Best run:", best_run, ":", dict_params[best_run]['loss'])

    return dict_params[best_run], dict_params if save_history else {}



# ------------------------------------------------------------------------------
# Project: IFE_Surrogate
# Authors: Tobias Leitgeb, Julian Tischler
# CD Lab 2025
# ------------------------------------------------------------------------------
import jax.numpy as jnp
from jax import vmap
from jaxtyping import Array

def nrms(test_y: Array, prediction: Array)-> Array:
        r"""
        Calculates the Normalized Root Mean Square Error.

        .. math:: 

            \text{NRMS} = \frac{\sqrt{\frac{1}{N}\sum_{i=1}^{N} (y_i - \hat{y}_i)^2}}{\max(y) - \min(y)}

        Args:
            test_y (ndarray): The true output values of shape (n, p).
            prediction (ndarray): The predicted output values of shape (n, p).

        Returns:
            float: The Root Mean Square Percentage Error.
        """
        test_y = jnp.atleast_2d(test_y)
        prediction = jnp.atleast_2d(prediction)

        assert test_y.shape == prediction.shape, "test_y and prediction must have the same shape"

        p = test_y.shape[1]
        error = (
            jnp.sqrt(1 / p * jnp.sum((test_y - prediction) ** 2, axis=1))
            / (jnp.max(test_y, axis=1) - jnp.min(test_y, axis=1)) 
        )
        return error

def rmse(test_y: Array, prediction: Array)-> Array:
        r"""
        Calculates the Root Mean Square Error.

        .. math:: 

            RMSPE = \sqrt{\frac{1}{n} \sum_{i=1}^{n} \left( \frac{y_i - \hat{y}_i}{y_i} \right)^2} 
        Args:
            test_y (ndarray): The true output values of shape (n, p).
            prediction (ndarray): The predicted output values of shape (n, p).

        Returns:
            Array: The Root Mean Square Error. 
        """
        test_y = jnp.atleast_2d(test_y)
        prediction = jnp.atleast_2d(prediction)

        assert test_y.shape == prediction.shape, "test_y and prediction must have the same shape"
        p = test_y.shape[1]
        error = jnp.sqrt(1 / p * jnp.sum(((test_y - prediction)/test_y) ** 2, axis=1))
        return error

        


# ------------------------------------------------------------------------------
# Project: IFE_Surrogate
# Authors: Tobias Leitgeb, Julian Tischler
# CD Lab 2025
# ------------------------------------------------------------------------------
from abc import ABC, abstractmethod
from typing import Callable, Dict, Tuple, List, Set
from jaxtyping import Array, Key
from flax import struct



@struct.dataclass
class Kernel(ABC):
    """
    Abstract base class for kernels in Gaussian Process (GP) regression.

    This class provides the core interface and utility functions for all kernel
    implementations, including methods to evaluate the kernel, access and
    update parameters, and sample new hyperparameters from prior distributions.

    Attributes:
        priors (Dict[str, Callable]): A dictionary mapping parameter names to 
            callable prior distributions used for sampling hyperparameters.
        param_bounds (Dict, optional): Dictionary specifying bounds for each 
            parameter. Defaults to an empty dictionary.
        exempt (Set[str]): Set of attribute names to exclude when retrieving 
            parameters (e.g., priors, param_bounds, exempt itself).

    Methods:
        evaluate(x1, x2):
            Abstract method. Computes the kernel function between two inputs.
        __call__(x1, x2):
            Calls `evaluate`, allowing the kernel object to be used as a function.
        get_priors():
            Returns the dictionary of priors for kernel hyperparameters.
        get_params():
            Returns a dictionary of the current kernel parameters, excluding
            those in `exempt`.
        sample_hyperparameters(key):
            Samples new parameter values from the prior distributions using
            a JAX random key.
        update_params(params):
            Updates kernel parameters in-place using the provided dictionary.
    """

    priors: Dict[str, Callable] = struct.field(pytree_node=False, kw_only=True)
    param_bounds: Dict = struct.field(pytree_node=False, default_factory=dict, kw_only=True)
    exempt: Set[str] = struct.field(
        pytree_node=False,
        default_factory=lambda: {'priors', 'param_bounds', 'exempt'},
        kw_only=True
    )
    
    @abstractmethod
    def evaluate(self, x1: Array, x2: Array) -> Array:
        """
        Evaluate the kernel function.

        Args:
            x1: First input data array.
            x2: Second input data array.

        Returns:
            Kernel evaluation between x1 and x2.
        """
        pass

    def __call__(self, x1: Array, x2: Array) -> Array:
        """
        Args:
            x1: First input data array.
            x2: Second input data array.

        Returns:
            Kernel evaluation between x1 and x2.
        """
        return self.evaluate(x1, x2)
    
    def get_priors(self):
        """
        Get the kernel parameters.

        Returns:
            A dictionary containing the kernel parameters.
        """
        return self.priors
    
    def get_params(self) -> Dict[str, Array]:
        """
        Get the kernel parameters.

        Returns:
            A dictionary containing the kernel parameters.
        """
        return {x: y for x, y in self.__dict__.items() if x not in self.exempt}
        

    def sample_hyperparameters(self, key: Key) -> Dict[str, Array]:
        """
        Sample new parameters from the prior distribution.

        Args:
            key: JAX random key.

        Returns:
            A dictionary containing sampled parameters.
        """
        assert isinstance(key, Array), "'key' must be a jax.random key"
        if self.priors is None:
            print("No priors specified. Returning current parameters.")
            return self.get_params()

        param_keys = self.get_params().keys()
        
        param_samples = {}
        for x in param_keys:
            if x in self.priors:
                param_samples[x] = self.priors[x].sample(key, self.__dict__[x].shape)
            else:
                param_samples[x] = self.__dict__[x]
        
        return param_samples


    def update_params(self, params: Dict[str, Array]) -> None:
        """
        Update the kernel parameters.

        Args:
            params: A dictionary containing the new parameters.
        """

        for key in params:
            self.__dict__[key] = params[key]

    



# ------------------------------------------------------------------------------
# Project: IFE_Surrogate
# Authors: Tobias Leitgeb, Julian Tischler
# CD Lab 2025
# ------------------------------------------------------------------------------
from dataclasses import dataclass
from jaxtyping import Array
from typing import Callable, Dict
import jax.numpy as jnp
from flax import struct
from .kernel import Kernel



@struct.dataclass
class Scale(Kernel):
    r"""
    Scale Kernel.

    The Scale kernel is a simple kernel that outputs a constant value across all
    pairs of points. It is often used in combination with other kernels to add
    a scaling factor.

    The kernel is defined as:

    .. math::

        k(\mathbf{x}_1, \mathbf{x}_2) = \sigma_f^2

    Args:
        variance (Array): The variance (scaling factor) for the kernel.
        priors (Dict[str, Callable], optional): Priors for the hyperparameter
            (variance).
    """
    variance: Array 
    # _priors: Dict[str, Callable] = struct.field(pytree_node=False)
    # _param_bounds: Dict = struct.field(pytree_node=False, default_factory=dict)

    def evaluate(self, x1: Array, x2: Array) -> Array:
        """
        Evaluate the Scale kernel between x1 and x2.

        This kernel returns a constant value, which is the variance, for all pairs of points in x1 and x2.

        Args:
            x1 (Array): Input array of shape (n_samples_1, n_features).
            x2 (Array): Input array of shape (n_samples_2, n_features).

        Returns:
            Array: Kernel matrix of shape (n_samples_1, n_samples_2), where each element is the `variance`.
        """
        return self.variance * jnp.ones((x1.shape[0], x2.shape[0]))



# ------------------------------------------------------------------------------
# Project: IFE_Surrogate
# Authors: Tobias Leitgeb, Julian Tischler
# CD Lab 2025
# ------------------------------------------------------------------------------

from jaxtyping import Array
from typing import Tuple
import jax.numpy as jnp
from jax import scipy as jsp
from jax import jit

from ife_surrogate.utils.decorators import experimental


def predictive_mean_var_wideband(
        X_train: Array, 
        Y_train: Array,
        sigma_sq: float,
        kernel ,
        jitter: float,
        X_test: Array,
    ) -> Tuple[Array, Array]:
    r"""
    Calculates the posterior mean and variance for a wideband (multi-output) 
    Gaussian Process with a diagonal noise covariance:
    
    .. math::
    
        p(\mathbf{f_{*,p}}|\mathbf{X},\mathbf{y_p},\mathbf{X_*})  =& \; \mathcal{N}(\mathbf{f_*}|\mathbf{\mu_{p}},\mathbf{\Sigma_p}) \\
        \mathbf{\mu_p}                          =&  \; \mathbf{K_*}^{T}\mathbf{K_x}^{-1} \mathbf{y_p} \\
        \mathbf{\Sigma_p}                       =& \; \sigma_p^2 (\mathbf{K_{**}} - \mathbf{K_*}^{T}\mathbf{K_x}^{-1}\mathbf{K_*})

    Args:
        X_train (Array): Training input data of shape (N, D).
        Y_train (Array): Training output data of shape (N, P).
        sigma_sq (float): Variance of the noise.
        kernel: Kernel function object used to compute covariances.
        jitter (float): Small value added to the diagonal for numerical stability.
        X_test (Array): Test input data of shape (M, D).

    Returns:
        Tuple[Array, Array]:  Posterior mean and variance at the test inputs for every task.
            - posterior mean: shape (M, P)
            - posterior variance: shape (M,)
    """
    
    N = X_train.shape[0]
    K = kernel(X_train, X_train) * sigma_sq
    L = jsp.linalg.cho_factor(K + jitter*jnp.eye(N), lower=True)
    alpha = jsp.linalg.cho_solve(L, Y_train)
    k_star = kernel(X_train, X_test) * sigma_sq
    f_star = jnp.dot(k_star.T, alpha)
    v = jnp.linalg.solve(L[0], k_star)
    var_f_star = kernel(X_test, X_test) * sigma_sq - jnp.dot(v.T, v)
    var_f_star = jnp.diag(var_f_star)
    return f_star, var_f_star



def predictive_mean_var_scalar(
        X_train: Array, 
        Y_train: Array,
        kernel ,
        jitter: float,
        X_test: Array,
    ) -> Tuple[Array, Array]:
    r"""
    Calculates the posterior mean and variance for a scalar output GP.
    Posterior distribution with posterior mean and variance:
    
    .. math::
    
        p(\mathbf{f_{*}}|\mathbf{X},\mathbf{y},\mathbf{X_*})  =& \; \mathcal{N}(\mathbf{f_*}|\mathbf{\mu},\mathbf{\Sigma}) \\
        \mathbf{\mu_p}                          =&  \; \mathbf{K_*}^{T}\mathbf{K}^{-1} \mathbf{y} \\
        \mathbf{\Sigma_p}                       =& \; \mathbf{K_{**}} - \mathbf{K_*}^{T}\mathbf{K}^{-1}\mathbf{K_*})    
    """
    N = X_train.shape[0]
    K = kernel(X_train, X_train)
    L = jsp.linalg.cho_factor(K + jitter*jnp.eye(N), lower=True)
    #predictive mean
    alpha = jsp.linalg.cho_solve(L, Y_train)
    k_star = kernel(X_train, X_test)
    f_star = k_star.T@alpha
    #predictive variance
    v = jnp.linalg.solve(L[0], k_star)
    cov_star = kernel(X_test, X_test) - jnp.dot(v.T, v)
    return f_star, jnp.diag(cov_star)



@experimental
def predictive_mean_var_kron(
        X_train: Array, 
        Y_train: Array,
        kernel,
        jitter: float,
        XW_test: Array,
    ) -> Tuple[Array, Array]:

    r"""
    Predictive mean for Kron GP with Kw = diag(alpha) Cw diag(alpha). Alpha will be linearly interpolated 
    for predictions. Where alpha test = func(f_test| f_train, std(Y_train))
    """

    N, P = Y_train.shape
    X, W = X_train
    X_test, W_test = XW_test
        # --- compute training alpha and interpolate to test tasks
    alpha_train = jnp.sqrt(jnp.var(Y_train, axis=0, ddof=1)).squeeze()  # shape (P,)
    alpha_test = jnp.interp(W_test, W.ravel(), alpha_train.ravel()).squeeze() # shape (P*,)

    
    K_x, C_w = kernel.evaluate((X, W), (X, W))   # return Kx, Cw
    K_w = (alpha_train[:, None] * C_w) * alpha_train[None, :]


    L_x = jnp.linalg.cholesky(K_x + jitter * jnp.eye(N))
    L_w = jnp.linalg.cholesky(K_w + jitter * jnp.eye(P))# * jnp.eye(P))

    # Step 1: Solve in space
    Kx_s = kernel.kernel_1(X_test, X)
    A = jsp.linalg.cho_solve((L_x, True), Y_train)       # N×P
    A_pred_space = Kx_s @ A                              # N* × P

    # --- Step 2: task solve
    Cw_s = kernel.kernel_2(W_test, W)                               # (P*, P) correlation part
    Kw_s = (alpha_test[:, None] * Cw_s) * alpha_train[None, :]      #scale with alpha

    B = jsp.linalg.cho_solve((L_w, True), A_pred_space.T)  # (P, N*)
    mu_star = (Kw_s @ B).T                         # (N*, P*)

    # predictive variance in two steps
    # space
    V_space = Kx_s @ jsp.linalg.cho_solve((L_x, True), Kx_s.T)  # N* x N*
    # tasks
    V_task = Kw_s @ jsp.linalg.cho_solve((L_w, True), Cw_s.T)  # P* x P*
    # Kronecker approximation: variance at each N* x P* element
    N_test, P_test = X_test.shape[0], W_test.shape[0]
    var_star  = jnp.diag(jnp.kron(V_space, V_task)).reshape(N_test, P_test)

    return mu_star, var_star


# ------------------------------------------------------------------------------
# Project: IFE_Surrogate
# Authors: Tobias Leitgeb, Julian Tischler
# CD Lab 2025
# ------------------------------------------------------------------------------
"""
Contains all Gaussian Process models of the Library.
"""
from .gp_model import GPModel
from .scalar_gp import ScalarGP, ScalarGPBaysian
from .wideband_gp import WidebandGP, WidebandGPBaysian
from .multi_output_gp import SeparableMultiOutputGP


__all__ = [
    "GPModel",
    "WidebandGP",  
    "WidebandGPBaysian",
    "ScalarGP",
    "ScalarGPBaysian",
    "SeparableMultiOutputGP",
    ]